import os
from textwrap import dedent


class Table:
    def __init__(self, duck_obj, table_name, force_index_join=False):
        self.duck = duck_obj
        self.table_name = table_name
        self._exists_dict = {}
        self.force_index_join = force_index_join

    def __dir__(self):  # pragma: no cover
        return [
            "df",
            "head",
            "insert",
        ]

    def ensure_writeable(self):
        if self.duck._read_only:
            raise ValueError("Trying to modify read-only database")

    def query(self, sql, fetch=True, **kwargs):
        return self.duck.query(sql, fetch=fetch, **kwargs)

    @property
    def df(self):
        df = self.query(f"SELECT * FROM {self.table_name}")
        return df

    def head(self, n=5):
        return self.query(f"SELECT * FROM {self.table_name} LIMIT {n}")

    def create(self, df):
        self.drop()
        self.query(
            f"BEGIN TRANSACTION; CREATE TABLE {self.table_name} AS SELECT * FROM __df_in__; COMMIT",
            fetch=False,
            __df_in__=df,
        )
        self.exists_dict = {}
        self.duck.connection.unregister("__df_in__")

    def insert(self, df):
        self.ensure_writeable()
        self.query(
            f"BEGIN TRANSACTION INSERT INTO {self.table_name} SELECT * FROM __df_in__ COMMIT",
            fetch=False,
            __df_in__=df,
        )
        self.duck.connection.unregister("__df_in__")

    def drop(self):
        self.ensure_writeable()
        self.query(f"DROP TABLE IF EXISTS {self.table_name}", fetch=False)


class Tables:
    def __init__(self, duck_obj):
        self.duck = duck_obj

    def __dir__(self):  # pragma: no cover
        return ["create", "drop", "drop_all"] + [
            t for t in self.duck.table_names if not t.startswith("__idx")
        ]

    def create(self, name, df):
        if not isinstance(name, str):
            raise ValueError("The first argument must be the name")

        if name == "create":
            raise ValueError(
                "You cannot name a table 'create'.  It is a reserved word."
            )

        table = Table(self.duck, name)
        table.create(df)

    def drop(self, name):
        table = getattr(self, name)
        table.drop()

    def drop_all(self):
        table_names = list(self.duck.table_names)
        for name in table_names:
            self.drop(name)

    def __getattr__(self, name):
        if name in self.duck.table_names:
            table = Table(self.duck, name)
            setattr(self, name, table)
        else:
            raise AttributeError(f"No attribute {name}")
        return table

    def __str__(self):  # pragma: no cover
        return repr(self.duck.table_names)

    def __repr__(self):  # pragma: no cover
        return self.__str__()


class DuckModel:
    """
    A class for working with duckdb

    print(ezr.DuckModel.example)

    ...
    """

    example = dedent(
        """
        # Use default file_name
        duck = Duck()

        # Add a table to the database
        duck.tables.create(table_name, df)

        # Insert data to existing table
        duck.tables.<table_name>.insert(df)

        # Drop a table
        duck.tables.drop(table_name)

        # Get a list of all table-names
        duck.table_names
        or
        duck.tables.<tab>

        # Run sql on database
        duck.query(sql, **kwargs) . # (see docstring for kwargs)

        # Drop all tables
        duck.tables.drop_all()
    """
    )

    def __init__(
        self,
        file_name="./duck.ddb",
        overwrite=False,
        read_only=False,
        force_index_join=False,
    ):
        self.file_name = file_name
        self._read_only = read_only
        self._force_index_join = force_index_join
        self._connection = None

        if overwrite and os.path.isfile(self.file_name):
            os.unlink(self.file_name)
        if overwrite and os.path.isfile(f"{self.file_name}.wal"):
            os.unlink(f"{self.file_name}.wal")

        if overwrite and read_only:
            raise ValueError(
                "It doesn't make sense to set read_only and overwrite at the same time"
            )

        self.reset_connection()
        self.tables = Tables(self)

    def __dir__(self):  # pragma: no cover
        return [
            "tables",
            "table_names",
            "query",
            "explain",
            "reset_connection",
            "set_index",
            "list_indexes",
            "file_name",
        ]

    def reset_connection(self):
        """
        This resets the connection to the database eliminating registered dataframes
        and defined indexes.
        """
        import duckdb

        self._connection = duckdb.connect(self.file_name, read_only=self._read_only)
        if self._force_index_join:
            self.query("PRAGMA force_index_join", fetch=False)

    @property
    def connection(self):
        if self._connection is None:
            self.reset_connection()
        return self._connection

    def drop_connection(self):
        """
        Drops the connection to the database.  If you are using
        ibis sometimes the ibis and DuckModel connections can fight.
        """
        if self._connection is not None:
            self._connection.close()
        self._connection = None

    def set_index(self, table_name, col_name):
        index_name = f"__idx_{table_name}_{col_name}__"
        self.query(
            f"create unique index {index_name} on {table_name} ({col_name})",
            fetch=False,
        )

    def list_indexes(self):
        return self.query("select * from pg_indexes")

    def explain(
        self, sql, **kwargs
    ):  # pragma: no cover because this is just used in debug
        sql = f"explain {sql}"
        res = self.query(sql, **kwargs)
        print(res.explain_value.iloc[0])

    def query(self, sql, fetch=True, **kwargs):
        """
        Run a sql query across the database.

        Args:
            sql: The sql you want to run
            fetch: If set to True (default) a dataframe will be returned.
                   Otherwise, the query will be executed without returning anything
            **kwargs: This is a tricky one.
                       DuckDB allows you to register dataframes as tables within a database
                       The kwargs to the query allow you to inject fake tables into the database
                       that hold the contents of as specified dataframe.

                       So, for example
                       duck = Duck()
                       duck.tables.create('one', df1)
                       duck.query("select * from one join two on my_id" two=df2)

        """
        for key, value in kwargs.items():
            self.connection.register(key, value)
        self.connection.execute(sql)
        if fetch:
            return self.connection.fetchdf()

    @property
    def table_names(self):
        df = self.query("PRAGMA show_tables")
        return sorted(df.name)

    def __del__(self):
        self.connection.close()



# Directly copied from:
# https://gist.githubusercontent.com/zichongkao/d2b2082181dd9628add017ee71a8a64f/raw/799949af9d3c9c24c223149f9270250113f181e2/nice_dates.py


# import pandas as pd
# import matplotlib.dates as mdates
# from datetime import timedelta as tdelta
# from matplotlib.ticker import FuncFormatter
# import matplotlib.pylab as plt

INTERVALS = {
    "YEARLY": [1, 2, 4, 5, 10],
    "MONTHLY": [1, 2, 3, 4, 6],
    "DAILY": [1, 2],
    "WEEKLY": [1, 2],
    "HOURLY": [1, 2, 3, 4, 6, 12],
    "MINUTELY": [1, 5, 10, 15, 30],
    "SECONDLY": [1, 5, 10, 15, 30],
}

YAXIS_WIDTH_INCH = 2.0
TICKS_PER_INCH = 1.5


def _next_largest(value, options):
    for i in options:
        if i >= value:
            return i
    return i


def _get_dynamic_formatter(timedelta, *fmts):
    def dynamic_formatter(x, pos):
        import matplotlib.dates as mdates

        dx = mdates.num2date(x)
        strparts = [dx.strftime(fmt) for fmt in fmts]
        if pos > 0:
            # renders previous tick and removes common parts
            prior_dx = dx - timedelta
            prior_strparts = [prior_dx.strftime(fmt) for fmt in fmts]
            strparts = [
                new if new != prior else ""
                for new, prior in zip(strparts, prior_strparts)
            ]
        return "\n".join(strparts).strip()

    return dynamic_formatter


def _deduce_locators_formatters(max_ticks, data):
    from datetime import timedelta as tdelta
    import matplotlib.dates as mdates
    from matplotlib.ticker import FuncFormatter

    data_interval_seconds = (data[-1] - data[0]) / tdelta(seconds=1)
    interval_seconds = data_interval_seconds / max_ticks

    if interval_seconds < tdelta(minutes=0.5).total_seconds():
        # print("xticks: seconds")
        unit_multiple = _next_largest(interval_seconds, INTERVALS["SECONDLY"])
        timedelta = tdelta(seconds=unit_multiple)
        return (
            mdates.SecondLocator(bysecond=range(0, 60, unit_multiple)),
            FuncFormatter(_get_dynamic_formatter(timedelta, "%M%:S", "%-Hh", "%-d %b")),
        )
    elif interval_seconds < tdelta(hours=0.5).total_seconds():
        # print("xticks: minutes")
        unit_multiple = _next_largest(
            interval_seconds / tdelta(minutes=1).total_seconds(), INTERVALS["MINUTELY"]
        )
        timedelta = tdelta(minutes=unit_multiple)
        return (
            mdates.MinuteLocator(byminute=range(0, 60, unit_multiple)),
            FuncFormatter(_get_dynamic_formatter(timedelta, "%H%:M", "%-d %b", "%Y")),
        )
    elif interval_seconds < tdelta(days=0.5).total_seconds():
        # print("xticks: hours")
        unit_multiple = _next_largest(
            interval_seconds / tdelta(hours=1).total_seconds(), INTERVALS["HOURLY"]
        )
        timedelta = tdelta(hours=unit_multiple)
        return (
            mdates.HourLocator(byhour=range(0, 24, unit_multiple)),
            FuncFormatter(_get_dynamic_formatter(timedelta, "%-Hh", "%-d %b", "%Y")),
        )
    elif interval_seconds < tdelta(days=3).total_seconds():
        # print("xticks: days")
        unit_multiple = _next_largest(
            interval_seconds / tdelta(days=1).total_seconds(), INTERVALS["DAILY"]
        )
        timedelta = tdelta(days=unit_multiple)
        return (
            mdates.WeekdayLocator(byweekday=range(0, 7, unit_multiple)),
            FuncFormatter(_get_dynamic_formatter(timedelta, "%-d", "%b", "%Y")),
        )
    elif interval_seconds < tdelta(days=14).total_seconds():
        # print("xticks: weeks")
        unit_multiple = _next_largest(
            interval_seconds / tdelta(weeks=1).total_seconds(), INTERVALS["WEEKLY"]
        )
        timedelta = tdelta(days=unit_multiple * 7)
        return (
            mdates.WeekdayLocator(byweekday=0, interval=unit_multiple),
            FuncFormatter(_get_dynamic_formatter(timedelta, "%-d", "%b", "%Y")),
        )
    elif interval_seconds < tdelta(weeks=26).total_seconds():
        # print("xticks: months")
        unit_multiple = _next_largest(
            interval_seconds / tdelta(weeks=4).total_seconds(), INTERVALS["MONTHLY"]
        )
        timedelta = tdelta(weeks=unit_multiple * 4)
        return (
            mdates.MonthLocator(bymonth=range(1, 13, unit_multiple)),
            FuncFormatter(_get_dynamic_formatter(timedelta, "%b", "%Y")),
        )
    else:
        # print("xticks: years")
        unit_multiple = _next_largest(
            interval_seconds / tdelta(weeks=52).total_seconds(), INTERVALS["YEARLY"]
        )
        return (mdates.YearLocator(base=unit_multiple), mdates.DateFormatter("%Y"))


def nice_dates(ax=None):
    import matplotlib.pylab as plt
    import pandas as pd

    if ax is None:
        fig = plt.gcf()
        ax = plt.gca()
    else:
        fig = ax.figure

    # information for deciding tick locations
    xaxis_length_inch = fig.get_size_inches()[0] * fig.dpi / 100
    max_ticks = (xaxis_length_inch - YAXIS_WIDTH_INCH) * TICKS_PER_INCH
    data = pd.to_datetime(ax.lines[0].get_xdata())

    maj_locator, maj_formatter = _deduce_locators_formatters(max_ticks, data)

    ax.xaxis.set_major_locator(maj_locator)
    ax.xaxis.set_major_formatter(maj_formatter)



#! /usr/bin/env python

import argparse
import pathlib


def parse_args():
    parser = argparse.ArgumentParser(
        description="Create a Makefile for running marimo notebooks"
    )
    parser.add_argument(
        "--notebook",
        "-n",
        help="Name of the notebook file",
    )
    return parser.parse_args()


MAKEFILE_TEMPLATE = """#! /usr/bin/make 
.PHONY: help
help:  ## Print the help documentation
	@grep -E '^[a-zA-Z_-]+:.*?## .*$$' $(MAKEFILE_LIST) | sort | awk 'BEGIN {{FS = ":.*?## "}}; {{printf "\\033[36m%-30s\\033[0m %s\\n", $$1, $$2}}'


.PHONY: run
run: ## Run the notebook as an app
	# marimo run --sandbox {notebook_name}
	uv run --with marimo marimo run --sandbox {notebook_name}

.PHONY: edit
edit: ## Run the notebook in edit mode
	uv run --with marimo marimo edit --sandbox  {notebook_name}
	# marimo edit --sandbox {notebook_name}
"""


def get_makefile(notebook_name):
    from fuzzypicker import picker

    if notebook_name is None:
        try:
            notebook_name = picker(
                ["notebook.nb.py"] + [f.name for f in pathlib.Path(".").glob("*.py")],
                default="notebook.nb.py",
            )
        except ImportError:
            return "notebook.nb.py"

    makefile_text = MAKEFILE_TEMPLATE.format(notebook_name=notebook_name)
    return makefile_text


def main():
    args = parse_args()
    with open("Makefile", "w") as f:
        f.write(get_makefile(args.notebook))


if __name__ == "__main__":
    main()



def tvd(y, lamda, num_iter=20, return_cost=False):
    """
    Performs total variation denoising on a supplied numpy array
    Args:
        lamda:
            The L1 regularization parameter.  A reaonable starting
            value for this is something like the standard devation of
            your noise.

        num_iter:
            This is an iterative algorithm.
            You may need to alter this number for good results

        return_cost:
            Set this to true if you also want to return a cost for each iteration
    Reurns:
        If return_cost is False
           - a numpy array of filtered results
        if return_ost is True
           (filtered_results, cost_array)
           - a numpy array of filtered results
           - a numpy array of lengh(num_iter) holding the cost at each iteration


    Implements algorithm written up by Ivan Selesnick.
    As of Sept. 2021 the paper was available from this url

    https://eeweb.engineering.nyu.edu/iselesni/lecture_notes/TVDmm/TVDmm.pdf

    It is also included in the references directory of this repository.
    """
    import numpy as np
    from scipy import sparse
    from scipy.sparse import linalg  # noqa weird scipy import thing

    if not isinstance(y, np.ndarray):
        raise ValueError("tvd can only be used on numpy arrays")

    if not len(y.shape) == 1:
        raise ValueError("tvd can only be used on flat arrays. Try y.flatten()")

    if np.any(np.isnan(y)):
        raise ValueError("tvd must not contain any nans")

    N = len(y)
    y = np.expand_dims(y, -1)
    cost = np.zeros(num_iter)
    I = sparse.identity(N, format="csr")  # noqa
    D = I[1:N, :] - I[: N - 1, :]
    DDT = D @ D.T

    x = y
    Dx = D @ x
    Dy = D @ y
    for k in range(num_iter):
        F = sparse.diags(np.abs(Dx.flatten()) / lamda, format="csr") + DDT
        x = y - np.expand_dims(D.T @ sparse.linalg.spsolve(F, Dy), -1)
        Dx = D @ x
        cost[k] = 0.5 * np.sum((x - y) ** 2) + lamda * np.sum(np.abs(Dx))

    if return_cost:
        return x.flatten(), cost
    else:
        return x.flatten()



from itertools import repeat, chain


def figure(*args, grid=True, style="default", figsize=(9, 5), **kwargs):
    """
    Returns a matplotlib axis object.
    """
    import pylab as pl

    available = [s for s in pl.style.available + ["default"] if not s.startswith("_")]
    if style not in available:
        raise ValueError(f"\n\n Valid Styles are {available}")
    pl.style.use(style)
    pl.figure(*args, figsize=figsize, **kwargs)
    pl.grid(grid)
    ax = pl.gca()
    return ax


class ColorCyle:
    def __init__(self):
        from string import ascii_lowercase
        import holoviews as hv

        hv.extension("bokeh")
        self.names, self.codes = zip(
            *(zip(ascii_lowercase, hv.Cycle().default_cycles["default_colors"]))
        )
        for name, code in zip(self.names, self.codes):
            setattr(self, name, code)

        self.reset()

    def reset(self):
        self._code_repeats = list(chain(*repeat(self.codes, 20)))
        self._code_iterator = iter(self._code_repeats)
        return self

    def __iter__(self):
        return self._code_iterator

    def __next__(self):
        return next(self._code_iterator)

    def __getitem__(self, index):
        return self._code_repeats[index]



import io
import os
import pandas as pd
import requests
import webbrowser
from .dataframe_tools import slugify as slugify_func
from typing import List, Optional


class SFDCEnv:
    def __init__(self):
        try:
            self.USERNAME = os.environ["SALESFORCE_USERNAME"]
            self.PASSWORD = os.environ["SALESFORCE_PASSWORD"]
            self.TOKEN = os.environ["SALESFORCE_SECURITY_TOKEN"]
        except KeyError:
            raise KeyError(
                (
                    "Can't find salesforce environment variables.  Required variables are\n"
                    "SALESFORCE_USERNAME\n"
                    "SALESFORCE_PASSWORD\n"
                    "SALESFORCE_SECURITY_TOKEN\n"
                )
            )


class SalesForceReport(SFDCEnv):
    def __init__(self):
        super().__init__()

        # Import here to avoid simple_salesforce dependency at ezr import
        import simple_salesforce

        self.session = requests.Session()
        self.sf_obj = simple_salesforce.Salesforce(
            username=self.USERNAME,
            password=self.PASSWORD,
            security_token=self.TOKEN,
            session=self.session,
        )

    def __del__(self):
        self.session.close()

    @classmethod
    def open_report(cls, report_id):
        webbrowser.open_new(
            f"https://ambition.lightning.force.com/lightning/r/Report/{report_id}/view"
        )

    def get_report(
        self,
        report_id: str,
        slugify: bool = False,
        date_fields: Optional[List[str]] = None,
        remove_copyright: bool = True,
    ):
        url = f"https://{self.sf_obj.sf_instance}/{report_id}"
        with requests.get(
            url,
            params=dict(export=1, xf="csv", enc="UTF-8", isdtp="p1"),
            cookies=dict(sid=self.sf_obj.session_id),
        ) as req:
            df = pd.read_csv(io.BytesIO(req.content))

        if remove_copyright:
            df = df.iloc[:-5, :]

        if slugify:
            df.columns = slugify_func(df.columns)

        if date_fields:
            for field in date_fields:
                df[field] = df[field].astype("datetime64[ns]")

        return df


class Soql(SFDCEnv):
    """

    # Use soql to query sfdc
    soql = Soql(sf_obj=None)
    df = sfb.query('SELECT Id FROM Order')

    # Access the underlying simple-salesorce api object to manage objects
    soql.sf.Order.update(df.Id.iloc[0], {'Name': 'My new name'})

    # Reference
    # https://simple-salesforce.readthedocs.io/en/latest/user_guide/record_management.html

    """

    def __init__(self, sf_obj=None):
        if sf_obj is None:
            # Import here to avoid simple_salesforce dependency at ezr import
            import simple_salesforce

            self.sf = simple_salesforce.Salesforce(
                username=self.USERNAME,
                password=self.PASSWORD,
                security_token=self.TOKEN,
            )
        else:
            self.sf = sf_obj

    def query(self, soql):
        rec_list = []
        for rec in self.sf.query(soql).get("records", []):
            rec.pop("attributes")
            rec_list.append(rec)

        df = pd.DataFrame(rec_list)
        return df



class Gemini:
    """
    A class for interacting with Google's Gemini AI models through the Vertex AI API.

    This class provides a simplified interface to generate text responses using Gemini models.
    It handles authentication, model configuration, and response generation in a streamlined way.

    ==================================================================================
    Note:  I just discovered the mirascope project.  Here are notes on how to do it.
        from mirascope import llm
        from pydantic import BaseModel
        gem = ezr.Gemini()

        class Response(BaseModel):
            my_message: str

        @llm.call(provider='google', model=gem.model, client=gem.client, response_model=Response)
        def doit(prompt):
            return prompt

        resp = doit('what is wikipedia')
        print(resp.my_message)
        print(resp._response.input_tokens)
        print(resp._response.output_tokens)
    ==================================================================================


    Attributes:
        COMPUTE_LOCATION (str): The Google Cloud compute location to use for API calls.
    """

    COMPUTE_LOCATION = "us-central1"

    def __init__(
        self,
        model="gemini-2.0-flash",
        temperature=0.01,
        top_p=0.95,
        max_output_tokens=8192,
        use_staging=True,
    ):
        """
        Initialize the Gemini client with specified configuration.

        Args:
            model (str, optional): The Gemini model to use. Defaults to "gemini-2.0-flash-001".
            temperature (float, optional): Controls randomness in the output. Lower values make
                the output more deterministic. Defaults to 0.01.
            top_p (float, optional): Controls diversity via nucleus sampling. Defaults to 0.95.
            max_output_tokens (int, optional): Maximum number of tokens to generate. Defaults to 8192.
            use_staging (bool, optional): Whether to use staging environment. Defaults to True.

        Note:
            Requires GOOGLE_CLOUD_PROJECT or GOOGLE_CLOUD_PROJECT_STAGING environment variables to be set.
        """
        import os
        from google import genai
        from google.genai import types

        self.model = model
        self.temperature = temperature
        self.top_p = top_p
        self.max_output_tokens = max_output_tokens

        if use_staging:
            project = os.environ["GOOGLE_CLOUD_PROJECT_STAGING"]
        else:
            project = os.environ["GOOGLE_CLOUD_PROJECT"]

        self.client = genai.Client(
            vertexai=True,
            project=project,
            location=self.COMPUTE_LOCATION,
        )

        self.config = types.GenerateContentConfig(
            temperature=self.temperature,
            top_p=self.top_p,
            max_output_tokens=max_output_tokens,
            response_modalities=["TEXT"],
            safety_settings=[
                types.SafetySetting(
                    category="HARM_CATEGORY_HATE_SPEECH", threshold="OFF"
                ),
                types.SafetySetting(
                    category="HARM_CATEGORY_DANGEROUS_CONTENT", threshold="OFF"
                ),
                types.SafetySetting(
                    category="HARM_CATEGORY_SEXUALLY_EXPLICIT", threshold="OFF"
                ),
                types.SafetySetting(
                    category="HARM_CATEGORY_HARASSMENT", threshold="OFF"
                ),
            ],
        )

    def prompt(self, text):
        """
        Generate a response from the Gemini model for the given input text.

        Args:
            text (str): The input text/prompt to send to the model.

        Returns:
            str: The generated response text from the model.

        Note:
            The response is streamed and concatenated before being returned.
        """
        from google.genai import types

        contents = [types.Content(role="user", parts=[types.Part.from_text(text=text)])]

        chunk_generator = self.client.models.generate_content_stream(
            model=self.model,
            contents=contents,
            config=self.config,
        )
        return ("".join([c.text for c in chunk_generator])).strip()



class Parallel:
    """
    A wrapper around native python parallel processing
    """

    def __init__(self, max_workers, kind="thread", show_progress=False):
        """
        Args:
            max_workers: The number of parallel jobs to run
            kind: "thread" or "process" determine how to parallelize
            show_progress: Display tqdm notebook-based progress bar

        Example:
            # Define the job you want to run
            def job(*args, **kwargs):
                import time
                time.sleep(args[0])
                return (args, kwargs)

            # Initialize the parallel processor
            p = ezr.Parallel(10)

            # Specify a list of args and kwargs you want to run
            call_tuples_list = [
                ([1], {'a': 1}),  # (args, kwargs)
                ([2], {'a': 2}),
                ([3], {'a': 3}),
            ]

            # Get results from a generator as jobs are completed
            for result in p.as_complete(job, call_tuples_list):
                print(result)
        """
        from concurrent import futures as futures_module

        self.futures_module = futures_module
        self.show_progress = show_progress

        if kind == "thread":
            self.executor = self.futures_module.ThreadPoolExecutor(
                max_workers=max_workers
            )

        elif kind == "process":
            import multiprocessing
            import warnings

            try:
                multiprocessing.set_start_method("fork")
            except RuntimeError:
                if multiprocessing.get_start_method() != "fork":
                    warnings.warn(
                        'multiprocessing.set_start_method("fork") raised an error.  You may need to restart kernel'
                    )

            self.executor = self.futures_module.ProcessPoolExecutor(
                max_workers=max_workers
            )
        else:
            allowed_kinds = ["thread", "process"]
            raise ValueError(f"kind must be one of {allowed_kinds}")

    def as_complete(self, func, call_tuple_iterable):
        """
        Yields results as they are completed. (This is a generator).
        Each parallel call takes the args tuple and kwargs dict
        specified in an element of the call_tuple_iterable.
        Each call tuple must contain a tuple or list followed by a dict

        p = Parallel(3)
        for result in p.as_complete(my_func, [(args1, kwargs1), ((args2, kwargs2)), ...]):
            print(result)
        """
        call_tuple_list = list(call_tuple_iterable)
        futures_obj_list = []
        for args, kwargs in call_tuple_list:
            futures_obj = self.executor.submit(func, *args, **kwargs)
            futures_obj_list.append(futures_obj)

        if self.show_progress:
            try:
                from tqdm.notebook import tqdm

                pbar = tqdm(total=len(call_tuple_list))
            except:  # noqa
                pass

        for results_obj in self.futures_module.as_completed(futures_obj_list):
            if self.show_progress:
                try:
                    pbar.update(1)
                except:  # noqa
                    pass
            yield results_obj.result()



import dataclasses
import typing


@dataclasses.dataclass
class Sample:
    name: str
    mu: float
    sigma: float
    sigma_of_mean: float = dataclasses.field(init=False)
    n: int
    dist: typing.Any = dataclasses.field(init=False)
    dist_of_mean: typing.Any = dataclasses.field(init=False)

    def __post_init__(self):
        from scipy import stats
        import numpy as np

        self.dist = stats.norm(loc=self.mu, scale=self.sigma)

        self.sigma_of_mean = self.sigma / np.sqrt(self.n)
        self.dist_of_mean = stats.norm(loc=self.mu, scale=self.sigma_of_mean)

        self.plot_min = self.mu - 5 * self.sigma
        self.plot_max = self.mu + 5 * self.sigma

        self.plot_min_of_mean = self.mu - 5 * self.sigma / np.sqrt(self.n)
        self.plot_max_of_mean = self.mu + 5 * self.sigma / np.sqrt(self.n)

    def plot(self, filled=True, use_uncertainty_of_mean=False):
        import numpy as np
        import holoviews as hv

        plot_points = 300
        if use_uncertainty_of_mean:
            dist = self.dist_of_mean
            xmin, xmax = self.plot_min_of_mean, self.plot_max_of_mean
        else:
            dist = self.dist
            xmin, xmax = self.plot_min, self.plot_max

        x = np.linspace(xmin, xmax, plot_points)
        y = dist.pdf(x)
        c = hv.Curve((x, y), label=self.name)
        if filled:
            c = hv.Area(c).options(alpha=0.15)
        return c


class NormalSampleJoiner:
    def __init__(self):
        import easier as ezr

        self.samples = ezr.Item()
        self.samples.combined = None

    def _check_args(self, data, mu, sigma, n):
        if data is None:
            if None in {mu, sigma, n}:
                raise ValueError(
                    "You must specify mu, sigma and n when you don't supply data"
                )
        else:
            if {None} != {mu, sigma, n}:
                raise ValueError(
                    "You cannot specify mu ,sigma or n when you supply data"
                )

    def __str__(self):
        if self.samples.combined is None:
            return "combined ~ (mu=None, sigma=None, n=None)"
        else:
            return (
                f"combined ~ (mu={self.samples.combined.mu:.2e}, "
                f"sigma={self.samples.combined.sigma:.2e}, n={self.samples.combined.n:.2e})"
            )

    def __repr__(self):
        return self.__str__()

    def _get_valid_name(self, name):
        if name == "combined":
            raise ValueError(
                "You cannot name a sample 'combined'.  That name is reserved."
            )

        if name is None:
            name = f"sample_{len(self.samples):03d}"
        return name

    def add_sample(self, *, data=None, mu=None, sigma=None, n=None, name=None):
        import numpy as np

        self._check_args(data, mu, sigma, n)
        name = self._get_valid_name(name)

        if data is not None:
            data = np.array(data)
            mu = np.mean(data)
            sigma = np.std(data)
            n = len(data)

        sample = Sample(
            name=name,
            mu=mu,
            sigma=sigma,
            n=n,
        )

        self._ingest(name, sample)

    def _ingest(self, name, sample):
        self.samples[name] = sample

        if self.samples.combined is None:
            self.samples.combined = sample
        else:
            new_combined = self.combine_samples(self.samples.combined, sample)
            self.samples.combined = new_combined

    def combine_samples(self, sample_a, sample_b, name="combined"):
        import numpy as np

        mu_a = sample_a.mu
        sigma_a = sample_a.sigma
        n_a = sample_a.n

        mu_b = sample_b.mu
        sigma_b = sample_b.sigma
        n_b = sample_b.n

        # Implement formulae in
        # https://github.com/robdmc/stats_for_software/blob/master/stats_for_software.pdf
        mu = n_a * mu_a / (n_a + n_b) + n_b * mu_b / (n_a + n_b)

        term1_numer = n_a * (sigma_a**2 + mu_a**2) + n_b * (sigma_b**2 + mu_b**2)
        term1_denom = n_a + n_b
        term2_numer = (n_a * mu_a + n_b * mu_b) ** 2
        term2_denom = (n_a + n_b) ** 2

        sigma2 = term1_numer / term1_denom - term2_numer / term2_denom
        sigma = np.sqrt(sigma2)

        n = n_a + n_b
        return Sample(name=name, mu=mu, sigma=sigma, n=n)

    def plot(self, use_uncertainty_of_mean=False, legend_position="top", filled=True):
        import holoviews as hv

        c_list = []
        for sample in self.samples.values():
            c_list.append(
                sample.plot(
                    filled=filled, use_uncertainty_of_mean=use_uncertainty_of_mean
                )
            )
        return hv.Overlay(c_list).options(legend_position=legend_position)



#######################################################################################################
# YOU MUST UPDATE THE TOOLS.PY FILE WHEN YOU ADD/REMOVE/CHANGE AN IMPORTABLE BY RUNNING cog -r tools.py
#######################################################################################################

import warnings


class Unsupported:
    def __init__(self, name):
        warnings.warn(
            f"THIS WARNING IS NOT CRITICAL: {name} is not supported without installing additional packages"
        )


class Importable:
    """
    This is a descriptor that knows how to load modules
    """

    def __init__(self, module_path, artifact_name, instantiate=False):
        self.instantiate = instantiate
        self._module_path = module_path
        self._artifact_name = artifact_name

    def __get__(self, instance, owner):
        # Import the module
        import importlib

        try:
            module = importlib.import_module(self._module_path)

            # Get the item (class/function/object) from the module
            thing = getattr(module, self._artifact_name)

            # If an object was requested, instantiate it
            if self.instantiate:
                thing = thing()
        except (ImportError, AttributeError, ModuleNotFoundError):
            thing = Unsupported(".".join([self._module_path, self._artifact_name]))

        return thing

    def __set__(self, instance, value):
        raise NotImplementedError("You cannot set this attribute")


class API:
    """
    Every object in the api is lazy loaded through a descriptor.  The actual
    attributes returned are classes/functions unless the instantiate flag is set.

    The lazy loading ensures that you only run imports if they are needed.
    """

    #######################################################################################################
    # YOU MUST UPDATE THE TOOLS.PY FILE WHEN YOU ADD/REMOVE/CHANGE AN IMPORTABLE BY RUNNING cog -r tools.py
    #######################################################################################################

    Animator = Importable("easier.hvtools", "Animator")
    Bernstein = Importable("easier.bernstein", "Bernstein")  # documented
    BernsteinFitter = Importable("easier.bernstein", "BernsteinFitter")  # documented
    BlobAttr = Importable("easier.utils", "BlobAttr")
    BlobMixin = Importable("easier.utils", "BlobMixin")
    Clock = Importable("easier.clock", "Clock")
    Crypt = Importable("easier.crypt", "Crypt")
    DistFitter = Importable("easier.distributions", "DistFitter")
    Duck = Importable("easier.duck", "Duck")
    # DuckModel = Importable("easier.duck_model", "DuckModel")
    DuckMirror = Importable("easier.duck_mirror", "DuckMirror")
    Elliptic = Importable("easier.filtering.elliptic_filter", "Elliptic")  # documented
    Fitter = Importable("easier.fit_lib", "Fitter")
    Gemini = Importable("easier.gemini", "Gemini")  # documented
    GlobalClock = Importable("easier.clock", "GlobalClock")
    Gsheet = Importable("easier.gsheet", "GSheet")  # documented
    Item = Importable("easier.item", "Item")
    Metabase = Importable("easier.metabase", "Metabase")
    MiniModel = Importable("easier.minimodel", "MiniModel")
    MiniModelPG = Importable("easier.minimodel", "MiniModelPG")
    MiniModelSqlite = Importable("easier.minimodel", "MiniModelSqlite")
    OrderedSchema = Importable(
        "easier.ibis_tools", "get_order_schema_class", instantiate=True
    )
    PG = Importable("easier.postgres", "PG")  # documented
    Parallel = Importable("easier.parallel", "Parallel")
    ParamState = Importable("easier.param_state", "ParamState")
    PrintCatcher = Importable("easier.print_catcher", "PrintCatcher")
    Proportion = Importable("easier.proportion_lib", "Proportion")
    QuickSchema = Importable(
        "easier.dataframe_tools", "get_quick_schema_class", instantiate=True
    )
    PandasSql = Importable(
        "easier.dataframe_tools", "get_pandas_sql_class", instantiate=True
    )
    SalesForceReport = Importable("easier.salesforce", "SalesForceReport")
    Scaler = Importable("easier.utils", "Scaler")
    Shaper = Importable("easier.shaper", "Shaper")
    Soql = Importable("easier.salesforce", "Soql")
    Timer = Importable("easier.timer", "Timer")
    beta_plots = Importable("easier.hvtools", "beta_plots")
    cached_container = Importable("easier.utils", "cached_container")  # documented
    cached_dataframe = Importable("easier.utils", "cached_dataframe")  # documented
    cached_property = Importable("easier.utils", "cached_property")  # documented
    cc = Importable("easier.plotting", "ColorCyle", instantiate=True)
    classifier_evaluation_plots = Importable(
        "easier.fit_lib", "classifier_evaluation_plots"
    )
    column_level_flattener = Importable(
        "easier.dataframe_tools", "column_level_flattener"
    )  # documented
    date_formatter = Importable("easier.nice_dates", "nice_dates")
    diff_strings = Importable("easier.utils", "diff_strings")
    django_reconnect = Importable("easier.utils", "django_reconnect")
    ecdf = Importable("easier.ecdf_lib", "ecdf")  # documented
    events_from_starting_ending = Importable(
        "easier.dataframe_tools", "events_from_starting_ending"
    )  # documented
    figure = Importable("easier.plotting", "figure")
    get_logger = Importable("easier.utils", "get_logger")
    heatmap = Importable("easier.dataframe_tools", "heatmap")  # documented
    hex_from_dataframe = Importable(
        "easier.dataframe_tools", "hex_from_dataframe"
    )  # documented
    hex_from_duckdb = Importable(
        "easier.dataframe_tools", "hex_from_duckdb"
    )  # documented
    hex_to_dataframe = Importable(
        "easier.dataframe_tools", "hex_to_dataframe"
    )  # documented
    hex_to_duckdb = Importable("easier.dataframe_tools", "hex_to_duckdb")  # documented
    hist = Importable("easier.hvtools", "hist")
    ibis_conn_to_sqlalchemy_conn = Importable(
        "easier.ibis_tools", "ibis_conn_to_sqlalchemy_conn"
    )
    ibis_duck_connection = Importable("easier.ibis_tools", "ibis_duck_connection")
    ibis_get_sql = Importable("easier.ibis_tools", "get_sql")
    ibis_postgres_connection = Importable(
        "easier.ibis_tools", "ibis_postgres_connection"
    )
    ibis_sql_to_frame = Importable("easier.ibis_tools", "sql_to_frame")
    in_notebook = Importable("easier.utils", "in_notebook")
    iterify = Importable("easier.iterify", "iterify")
    kill_outliers_iqr = Importable(
        "easier.outlier_tools", "kill_outliers_iqr"
    )  # documented
    kill_outliers_sigma_edit = Importable(
        "easier.outlier_tools", "kill_outliers_sigma_edit"
    )  # documented
    localize_utc_to_timezone = Importable(
        "easier.dataframe_tools", "localize_utc_to_timezone"
    )
    lomb_scargle = Importable("easier.lomb_scargle", "lomb_scargle")
    mem_get = Importable("easier.memory", "mem_get")
    mem_show = Importable("easier.memory", "mem_show")
    month_string = Importable("easier.dataframe_tools", "month_string")
    mute_warnings = Importable("easier.utils", "mute_warnings")
    nice_dates = Importable("easier.nice_dates", "nice_dates")
    NormalSampleJoiner = Importable("easier.normal_sampler", "NormalSampleJoiner")
    pandas_time_to_utc_seconds = Importable(
        "easier.dataframe_tools", "pandas_time_to_utc_seconds"
    )
    pandas_utc_seconds_to_time = Importable(
        "easier.dataframe_tools", "pandas_utc_seconds_to_time"
    )
    pg_creds_from_env = Importable("easier.postgres", "pg_creds_from_env")  # documented
    pickle_cache_mixin = Importable("easier.utils", "pickle_cache_mixin")  # documented
    pickle_cache_state = Importable("easier.utils", "pickle_cache_state")  # documented
    pickle_cached_container = Importable(
        "easier.utils", "pickle_cached_container"
    )  # documented
    print_error = Importable("easier.utils", "print_error")
    python_type = Importable("easier.utils", "python_type")
    screen_width_full = Importable("easier.utils", "screen_width_full")
    slugify = Importable("easier.dataframe_tools", "slugify")  # documented
    sql_file_to_df = Importable("easier.postgres", "sql_file_to_df")  # documented
    sql_string_to_df = Importable("easier.postgres", "sql_string_to_df")  # documented
    StreamPlotter = Importable("easier.stream_plotter", "StreamPlotter")
    tqdm_flex = Importable("easier.utils", "tqdm_flex")
    Tracker = Importable("easier.tracker", "Tracker")
    tvd = Importable("easier.filtering.tvd", "tvd")  # documented
    VonMisesFitter = Importable("easier.vonmises", "VonMisesFitter")
    warnings_mute = Importable("easier.utils", "mute_warnings")
    weekday_string = Importable("easier.dataframe_tools", "weekday_string")

    #######################################################################################################
    # YOU MUST UPDATE THE TOOLS.PY FILE WHEN YOU ADD/REMOVE/CHANGE AN IMPORTABLE BY RUNNING cog -r tools.py
    #######################################################################################################



from collections import namedtuple
from typing import List, Dict, Any, Tuple
import copy
import functools
import importlib
import os
import urllib.parse


def pg_creds_from_env(kind="dict", force_docker=False):
    """
    Pulls PostgreSQL credentials from environment variables or defaults to Docker settings.

    This function retrieves database connection credentials from environment variables.
    If environment variables are not set, it defaults to standard Docker PostgreSQL
    credentials. The credentials can be returned either as a dictionary or as a
    connection URL.

    Args:
        kind (str, optional): The format to return credentials in. Must be either
            "dict" or "url". Defaults to "dict".
        force_docker (bool, optional): If True, ignores environment variables and
            returns Docker default credentials. Defaults to False.

    Returns:
        Union[dict, str]: If kind="dict", returns a dictionary with connection
            parameters. If kind="url", returns a PostgreSQL connection URL string.

    Raises:
        ValueError: If kind is not one of the allowed values ("dict" or "url").
    """
    allowed_kinds = ["url", "dict"]

    if kind not in allowed_kinds:
        raise ValueError(f"Allowed kinds are {allowed_kinds}")

    if force_docker:
        env = {}
    else:
        env = os.environ.copy()

    creds = {
        "host": env.get("PGHOST", "db"),
        "port": env.get("PGPORT", "5432"),
        "database": env.get("PGDATABASE", "postgres"),
        "user": env.get("PGUSER", "postgres"),
        "password": env.get("PGPASSWORD", "postgres"),
    }
    # URL encode username and password to handle special characters
    encoded_user = urllib.parse.quote(creds["user"])
    encoded_password = urllib.parse.quote(creds["password"])
    url = f"postgresql://{encoded_user}:{encoded_password}@{creds['host']}:{creds['port']}/{creds['database']}"

    if kind == "dict":
        return creds
    else:
        return url


class PG:
    """
    A PostgreSQL database connection and query execution class.

    This class provides a convenient interface for connecting to PostgreSQL databases,
    executing queries, and retrieving results in various formats (DataFrame, tuples,
    dictionaries, etc.). It supports both direct connection parameters and Django
    database configurations.

    Args:
        host (str, optional): Database host. Defaults to PGHOST environment variable.
        user (str, optional): Database user. Defaults to PGUSER environment variable.
        password (str, optional): Database password. Defaults to PGPASSWORD environment variable.
        dbname (str, optional): Database name. Defaults to PGDATABASE environment variable.
        use_django (bool, optional): If True, use Django database settings. Defaults to False.

    Attributes:
        _sql (str): The SQL query to be executed
        _context (dict): Context for query execution
        _conn_kwargs (dict): Database connection parameters
        _raw_results (list): Raw query results
        _raw_columns (list): Column names from query results
    """

    _sql = None
    _context = None
    _conn_kwargs = None
    _raw_results = None
    _raw_columns = None

    def __init__(self, **kwargs):
        """
        Initialize a PostgreSQL database connection.

        Args:
            **kwargs: Connection parameters that can include:
                - host: Database host (default from PGHOST env var)
                - user: Database user (default from PGUSER env var)
                - password: Database password (default from PGPASSWORD env var)
                - dbname: Database name (default from PGDATABASE env var)
                - use_django: If True, use Django database settings (default: False)

        Using with Django
            After django is loaded, simply construct with

                pg = PG(use_django=True)

            Tables can be shown with
                pg.table_names()

            Django queries can be run with one line
            df = PG(use_django=True).query('SELECT * FROM my_table').df
        """
        # See if db info should be loaded from django
        use_django = kwargs.get("use_django", False)

        # If so, replace kwargs with values from django settings
        if use_django:
            self.safe_import("django")
            from django.conf import settings

            db = settings.DATABASES["default"]
            kwargs = {
                "host": db["HOST"],
                "user": db["USER"],
                "password": db["PASSWORD"],
                "dbname": db["NAME"],
            }

        env_translator = {
            "host": "PGHOST",
            "user": "PGUSER",
            "password": "PGPASSWORD",
            "dbname": "PGDATABASE",
        }
        conn_kwargs = {
            key: kwargs.get(key, os.environ.get(env_translator[key], None))
            for key in env_translator.keys()
        }

        bad_keys = []
        for key in conn_kwargs.keys():
            if conn_kwargs[key] is None:
                bad_keys.append(key)
        if bad_keys:
            raise ValueError(
                f"The following connections params not specified {bad_keys}"
            )

        self._conn_kwargs = conn_kwargs

    @classmethod
    def queryset_to_sql(cls, queryset):
        """
        Transform a Django queryset into formatted SQL that can be used in pg-admin.

        Args:
            queryset: A Django QuerySet object

        Returns:
            str: Formatted SQL query string
        """
        # Do imports here to avoid dependencies
        sqlparse = cls.safe_import("sqlparse")
        cls.safe_import("django")
        from django.db import connection

        # Compile the query to python db api
        sql, sql_params = queryset.query.get_compiler(using=queryset.db).as_sql()

        # Translate the python query spec into a postgres query
        with connection.cursor() as cur:
            query = cur.mogrify(sql, sql_params)

        # Make the query pretty and return it
        query = sqlparse.format(query, reindent=True, keyword_case="upper")
        return query

    def schema_names(self):
        """
        Get a list of all schema names in the database.

        Returns:
            pandas.DataFrame: DataFrame containing schema names
        """
        # Run schema query in a copied version of self so as not to mess with current query
        # on this object
        pg = copy.deepcopy(self)
        return pg.query("SELECT nspname FROM pg_catalog.pg_namespace").to_dataframe()

    @functools.lru_cache()
    def table_names(self, schema_name="public"):
        """
        Get a list of table names in the specified schema.

        Args:
            schema_name (str): Name of the schema to query (default: "public")

        Returns:
            pandas.DataFrame: DataFrame containing table names, sorted alphabetically
        """
        # Run table query in a copied version of self so as not to mess with current query
        # on this object
        pg = copy.deepcopy(self)
        df = pg.query(
            f"SELECT table_name FROM information_schema.tables WHERE table_schema='{schema_name}'"
        ).to_dataframe()
        df = df.sort_values(by="table_name")
        return df

    def query(self, sql) -> "PG":
        """
        Set the SQL query to be executed.

        Args:
            sql (str): SQL query string

        Returns:
            PG: Self instance for method chaining
        """
        self._sql = sql
        return self

    def run(self) -> "PG":
        """
        Execute the current SQL query and store results.

        Returns:
            PG: Self instance for method chaining

        Raises:
            psycopg2.ProgrammingError: If there's an error executing the query
        """
        psycopg2 = self.safe_import("psycopg2")
        with psycopg2.connect(**self._conn_kwargs) as connection:
            with connection.cursor() as cursor:
                cursor.execute(self._sql)
                try:
                    self._raw_results = list(cursor.fetchall())
                    self._raw_columns = [col[0] for col in cursor.description]
                except psycopg2.ProgrammingError as e:
                    if str(e) == "no results to fetch":
                        self._raw_results = []
                        self._raw_columns = []
                    else:  # pragma: no cover  No expected to hit this, but raise just in case
                        raise
        return self

    def reset(self):
        """
        Reset the query results and columns to None.
        """
        self._raw_columns = None
        self._raw_results = None

    @property
    def _results(self) -> List[Tuple]:
        """
        Get the raw query results, executing the query if necessary.

        Returns:
            List[Tuple]: List of result tuples
        """
        if self._raw_results is None:
            self.run()
        return self._raw_results

    @property
    def columns(self) -> List[str]:
        """
        Get the column names from the query results, executing the query if necessary.

        Returns:
            List[str]: List of column names
        """
        if self._raw_columns is None:
            self.run()
        return self._raw_columns

    def as_tuples(self) -> List[Tuple]:
        """
        Get query results as a list of tuples.

        Returns:
            List[Tuple]: List of result tuples
        """
        self.reset()
        return self._results

    def as_dicts(self) -> List[Dict[str, Any]]:
        """
        Get query results as a list of dictionaries.

        Returns:
            List[Dict[str, Any]]: List of dictionaries with column names as keys
        """
        self.reset()
        return [dict(zip(self.columns, row)) for row in self._results]

    def as_named_tuples(self, named_tuple_name="Result") -> List[Any]:
        """
        Get query results as a list of named tuples.

        Args:
            named_tuple_name (str): Name for the named tuple class (default: "Result")

        Returns:
            List[Any]: List of named tuples
        """
        self.reset()
        # Ignore typing in here because of unconventional namedtuple usage
        nt_result = namedtuple(named_tuple_name, self.columns)  # type: ignore
        return [nt_result(*row) for row in self._results]  # type: ignore

    @classmethod
    def safe_import(cls, module_name, package=None):
        """
        Safely import a module, raising a helpful error if import fails.

        Args:
            module_name (str): Name of the module to import
            package (str, optional): Package name for relative imports

        Returns:
            module: The imported module

        Raises:
            ImportError: If the module cannot be imported
        """
        try:
            imported = importlib.import_module(module_name, package=package)
        except (
            ImportError
        ):  # pragma: no cover.  Not going to uninstall pandas to test this
            raise ImportError(
                f"\n\nNope! This method requires that {module_name} be installed.  You know what to do."
            )

        return imported

    def as_dataframe(self) -> Any:
        """
        Get query results as a pandas DataFrame.

        Returns:
            pandas.DataFrame: DataFrame containing query results
        """
        self.reset()
        pd = self.safe_import("pandas")

        return pd.DataFrame(self._results, columns=self.columns)

    def to_tuples(self) -> List[Tuple]:
        """
        Alias for as_tuples().

        Returns:
            List[Tuple]: List of result tuples
        """
        return self.as_tuples()

    def to_dicts(self) -> List[Dict[str, Any]]:
        """
        Alias for as_dicts().

        Returns:
            List[Dict[str, Any]]: List of dictionaries with column names as keys
        """
        return self.as_dicts()

    def to_named_tuples(self) -> List[Any]:
        """
        Alias for as_named_tuples().

        Returns:
            List[Any]: List of named tuples
        """
        return self.as_named_tuples()

    def to_dataframe(self) -> Any:
        """
        Alias for as_dataframe().

        Returns:
            pandas.DataFrame: DataFrame containing query results
        """
        return self.as_dataframe()

    @property
    def df(self):
        """
        Property alias for to_dataframe().

        Returns:
            pandas.DataFrame: DataFrame containing query results
        """
        return self.to_dataframe()

    @property
    def sql(self) -> str:
        """
        Get the formatted SQL query string.

        Returns:
            str: Formatted SQL query string
        """
        sqlparse = self.safe_import("sqlparse")
        psycopg2 = self.safe_import("psycopg2")
        with psycopg2.connect(**self._conn_kwargs) as connection:
            with connection.cursor() as cursor:
                sql, params = self._get_prepared_query()
                query = cursor.mogrify(sql, params)
                query = sqlparse.format(query, reindent=True, keyword_case="upper")
        return query


def sql_file_to_df(file_name="sql_query.sql", context_dict=None):
    """
    Load and execute a SQL query from a file, returning the results as a DataFrame.
    Always connects to the postgres database with credentials from the environment.

    Args:
        file_name (str): Path to the SQL file. Defaults to 'sql_query.sql'.
        context_dict (dict, optional): Dictionary of variables to use for template rendering.
            Defaults to None.

    Returns:
        pandas.DataFrame: Results of the SQL query.
    """
    from pathlib import Path
    import jinja2
    import easier as ezr

    if context_dict is None:
        context_dict = {}

    # Expand user directory and get absolute path
    file_path = Path(file_name).expanduser().resolve()

    # Extract the directory the file lives in
    dir_path = file_path.parent

    jinja_env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(dir_path.as_posix()),
        autoescape=jinja2.select_autoescape(),
    )
    template = jinja_env.get_template(file_path.name)
    text = template.render(**context_dict)

    pg = ezr.PG()
    pg.query(text)
    df = pg.to_dataframe()
    return df


def sql_string_to_df(query, context_dict=None):
    """
    Execute a SQL query string with Jinja templating, returning the results as a DataFrame.
    Always connects to the postgres database with credentials from the environment.

    Args:
        query (str): SQL query string with optional Jinja template variables.
        context_dict (dict, optional): Dictionary of variables to use for template rendering.
            Defaults to None.

    Returns:
        pandas.DataFrame: Results of the SQL query.
    """
    import jinja2
    import easier as ezr

    if context_dict is None:
        context_dict = {}

    # Create a Jinja environment with a string loader
    jinja_env = jinja2.Environment(
        loader=jinja2.BaseLoader(), autoescape=jinja2.select_autoescape()
    )

    # Create a template from the query string
    template = jinja_env.from_string(query)

    # Render the template with the provided context
    rendered_query = template.render(**context_dict)

    # Execute the query and return results as DataFrame
    pg = ezr.PG()
    pg.query(rendered_query)
    df = pg.to_dataframe()
    return df



# flake8: noqa
from .api import API, Importable

# This object is loaded with descriptors that lazy import items from modules
api_obj = API()


class LazyTools:
    def __init__(self):
        for att_name, att_obj in API.__dict__.items():
            if isinstance(att_obj, Importable):
                setattr(self, att_name, getattr(api_obj, att_name))


lazy_tools = LazyTools()


# This list will hold all the items the package will export
__all__ = []

# We cycle over the api items and add their names to the exportable namespace
for att_name, att_obj in lazy_tools.__dict__.items():
    __all__.append(att_name)
    cmd = f"{att_name} = getattr(lazy_tools, '{att_name}')"
    exec(cmd)


# (I)Python looks at this function to determine what is exported
def __dir__():
    return __all__



import io
import os
import json


class Metabase:
    VALID_TYPES = ["category", "date", "datetime", "number", "text"]

    def __init__(self):
        # TODO: change the url to be read from the environment
        self.url_template = "https://metabase.app.blueberrypediatrics.com/api/card/{question_id}/query/csv?format_rows=true"
        self.headers = {
            "x-api-key": os.environ["METABASE_KEY"],
        }

    def _make_params_payload(self, **kwargs):
        parameters = {"parameters": []}

        for key, val in kwargs.items():
            try:
                param_type, param_value = val
            except:
                msg = "Kwargs error.  Must be like parameter_name=('cateory|date|datetime|number|text', parameter_value)"
                raise ValueError(msg)
            if param_type not in self.VALID_TYPES:
                raise ValueError(
                    f"Invalid type {param_type!r}.  Must be one of {self.VALID_TYPES}"
                )
            rec = {
                "type": param_type,
                "target": ["variable", ["template-tag", str(key)]],
                "value": [str(param_value)],
            }
            parameters["parameters"].append(rec)
        parameters["parameters"] = json.dumps(parameters["parameters"])
        return parameters

    def download_question(self, question_id, **kwargs):
        """
        Args:
            question_id: int = The question id you get from the metabase url
            kwargs = {'variable_name': ('cateory|date|datetime|number|text'), variable_value)}
        """
        import pandas as pd
        import requests

        url = self.url_template.format(question_id=question_id)
        params_dict = self._make_params_payload(**kwargs)
        if params_dict["parameters"]:
            post_kwargs = {"data": params_dict}
        else:
            post_kwargs = {}
        response = requests.post(url, headers=self.headers, **post_kwargs)
        if response.status_code != 200:
            raise ValueError(response.text)

        with io.StringIO(response.text) as buff:
            df = pd.read_csv(buff, low_memory=False)

        return df



import itertools
import json
import os
from textwrap import dedent
import string
from typing import Optional
from .utils import cached_property

import pandas as pd
import numpy as np
import datetime

# Setting the config file name like this is ugly.  Need a better way.
config_file_locations = [
    os.path.expanduser("/developer/.config/gspread/service_account.json"),
    os.path.expanduser("~/.config/gspread/service_account.json"),
]
for CONFIG_FILE_NAME in config_file_locations:
    if os.path.isfile(CONFIG_FILE_NAME):
        break


class Email:
    def __init__(self, config_file_name):
        self.config_file_name = config_file_name

    @cached_property
    def email(self):
        try:
            with open(self.config_file_name) as buff:
                blob = json.load(buff)

            return blob["client_email"]
        except:  # noqa
            return "email_not_found"

    def __get__(self, obj, objclass):
        return self.email

    def __set__(self, obj, val):
        raise NotImplementedError("Can't set the email property")


class Example:
    def __get__(self, obj, objclass):
        example = dedent(
            f"""
            Share document with this email address:

                {objclass.email}

            Then:
            goog = ezr.GSheet(document_name, sheet_name)
            goog.store_frame(edf)

        """
        )
        return example

    def __set__(self, obj, val):
        raise NotImplementedError("Can't set the example property")


class GSheet:
    """
    A class for interacting with Google Sheets using the Google Sheets API.

    This class provides methods to read from and write to Google Sheets,
    including functionality to work with pandas DataFrames.

    Attributes:
        email (str): The email address associated with the service account.
        example (str): An example showing how to use the GSheet class.
    """

    email = Email(CONFIG_FILE_NAME)
    example = Example()

    def __init__(self, doc, sheet):
        """
        Initialize a GSheet instance.

        Args:
            doc (str): The name or ID of the Google Sheets document.
            sheet (str): The name of the worksheet within the document.

        Raises:
            RuntimeError: If the service account configuration file is not found.
        """
        self._api = None
        self._document = None
        self._sheet = None
        self._df_cache = {}
        self.document = self.api.open(doc)
        self.sheet = self.document.worksheet(sheet)

        if not os.path.isfile(CONFIG_FILE_NAME):
            url = "http://gspread.readthedocs.io/en/latest/oauth2.html"
            msg = f"\n\nFile does not exist:\n{CONFIG_FILE_NAME}\n\nSee {url}"
            raise RuntimeError(msg)

    @classmethod
    def print_example(cls):
        """
        Print an example showing how to use the GSheet class.
        """
        print(cls.example)

    @cached_property
    def col_to_num(self):
        """
        Create a mapping from spreadsheet column names to column numbers.

        Returns:
            dict: A dictionary mapping column names (e.g., 'A', 'B', 'AA') to their
                  corresponding column numbers (1-based).
        """
        uppers = list(string.ascii_uppercase)
        return {
            val: ind + 1
            for (ind, val) in enumerate(
                "".join(list(t)) for t in list(itertools.product([""] + uppers, uppers))
            )
        }

    @cached_property
    def num_to_col(self):
        """
        Create a mapping from column numbers to spreadsheet column names.

        Returns:
            dict: A dictionary mapping column numbers (1-based) to their
                  corresponding column names (e.g., 'A', 'B', 'AA').
        """
        return {num: col for (col, num) in self.col_to_num.items()}

    @property
    def api(self):
        """
        Get the gspread API client.

        Returns:
            gspread.Client: An authenticated gspread client instance.
        """
        # Import here to avoid gspread dependency
        import gspread

        gc = gspread.service_account(CONFIG_FILE_NAME)
        return gc

    def to_dataframe_as_values(self):
        """
        Convert the entire sheet to a pandas DataFrame without header processing.

        Returns:
            pd.DataFrame: A DataFrame containing all values from the sheet.
        """
        lol = self.sheet.get_all_values()
        return pd.DataFrame(lol)

    def to_dataframe(self, header_row=1, reload=True):
        """
        Convert the sheet to a pandas DataFrame using the specified row as headers.

        Args:
            header_row (int, optional): The row number to use as column headers.
                                      Defaults to 1.
            reload (bool, optional): Whether to force reload the data from the sheet.
                                   Defaults to True.

        Returns:
            pd.DataFrame: A DataFrame containing the sheet data with proper column headers.
        """
        # get a key to identify this frame in the cache
        key = (self.document.title, self.sheet.title)

        # fill cache if needed and returned cached frame
        df = self._df_cache.get(key)
        if df is None or reload:
            recs = self.sheet.get_all_records(head=header_row)
            columns = self.sheet.row_values(header_row)
            self._df_cache[key] = pd.DataFrame(recs)[columns]
        return self._df_cache[key].copy()

    def read_cell(self, coord):
        """
        Read the value of a specific cell.

        Args:
            coord (str): The cell coordinate (e.g., 'A1', 'B2').

        Returns:
            Any: The value of the specified cell.
        """
        cell = self.sheet.acell(coord, value_render_option="UNFORMATTED_VALUE")
        return cell.value

    def write_cell(self, coord, value):
        """
        Write a value to a specific cell.

        Args:
            coord (str): The cell coordinate (e.g., 'A1', 'B2').
            value (Any): The value to write to the cell.
        """
        self.write_cells([(coord, value)])

    def read_formula(self, coord):
        """
        Read the formula of a specific cell.

        Args:
            coord (str): The cell coordinate (e.g., 'A1', 'B2').

        Returns:
            str: The formula in the specified cell.
        """
        cell = self.sheet.acell(coord, value_render_option="FORMULA")
        return cell.value

    def write_formula(self, coord, value):
        """
        Write a formula to a specific cell.

        Args:
            coord (str): The cell coordinate (e.g., 'A1', 'B2').
            value (str): The formula to write to the cell.
        """
        self.sheet.update_acell(coord, value)

    def write_cells(self, tuples):
        """
        Write multiple values to multiple cells.

        Args:
            tuples (list): A list of (coord, value) tuples where coord is a cell
                          coordinate (e.g., 'C15') and value is the value to write.
        """
        cells = []
        for coord, val in tuples:
            cell = self.sheet.acell(coord)
            cell.value = val
            cells.append(cell)

        self.sheet.update_cells(cells)

    def store_frame_to_coords(
        self,
        df: pd.DataFrame,
        top_left_coord: str,
        clear_to_bottom: bool = False,
        max_num_rows: Optional[int] = None,
        max_num_cols: Optional[int] = None,
    ):
        """
        Store a DataFrame to a specific location in the sheet.

        Args:
            df (pd.DataFrame): The DataFrame to store.
            top_left_coord (str): The spreadsheet coordinate (e.g., 'A15') of the top left corner.
            clear_to_bottom (bool, optional): Whether to clear all cells below the data.
                                            Defaults to False.
            max_num_rows (int, optional): Limit the number of rows to write.
                                        Defaults to None.
            max_num_cols (int, optional): Limit the number of columns to write.
                                        Defaults to None.
        """
        top_left_coord = top_left_coord.upper()

        df = df.reset_index(drop=True).fillna("")

        if max_num_rows:
            df = df.iloc[:max_num_rows, :]
        if max_num_cols:
            df = df.iloc[:, :max_num_cols]

        cell = self.sheet.acell(top_left_coord)
        top_row = cell.row
        left_col = cell.col

        bottom_row = cell.row + len(df)
        right_col = cell.col + len(df.columns) - 1

        if clear_to_bottom:
            range_string = (
                f"'{self.sheet.title}'!{top_left_coord}:{self.num_to_col[right_col]}"
            )
            self.document.values_clear(range_string)
        else:
            range_string = f"'{self.sheet.title}'!{top_left_coord}:{self.num_to_col[right_col]}{bottom_row}"
            self.document.values_clear(range_string)

        cells = self.sheet.range(top_row, left_col, bottom_row, right_col)

        gsheet_epoc = datetime.datetime(1899, 12, 30)
        for cell in cells:
            frame_row = cell.row - top_row - 1
            frame_col = cell.col - left_col
            if cell.row == top_row:
                cell.value = list(df.columns)[frame_col]
            else:
                value = df.values[frame_row, frame_col]
                # This is a hack.  Dates will be exported as days from epoch
                # Format this as date to get the dates you want
                if isinstance(value, pd.Timestamp):
                    value = (value.to_pydatetime() - gsheet_epoc).days
                elif value == "":
                    value = None
                elif type(value) in [np.int64, np.int32, np.int16, np.int8]:
                    value = int(value)
                cell.value = value
        self.sheet.update_cells(cells)

    def store_frame(self, df, starting_row=1, total_rows=None, total_cols=None):
        """
        Clear the sheet and store a DataFrame starting from the specified row.

        Args:
            df (pd.DataFrame): The DataFrame to store.
            starting_row (int, optional): The row number to start writing from.
                                        Defaults to 1.
            total_rows (int, optional): The total number of rows to allocate.
                                      Defaults to None.
            total_cols (int, optional): The total number of columns to allocate.
                                      Defaults to None.
        """
        df = df.reset_index(drop=True).fillna("")
        self.sheet.clear()

        if total_rows is None:
            needed_row_count = len(df) + 20
        else:
            needed_row_count = total_rows

        if total_cols is None:
            needed_col_count = len(df.columns) + 20
        else:
            needed_col_count = total_cols

        self.sheet.resize(rows=needed_row_count, cols=needed_col_count)

        min_row = starting_row
        max_row = min_row + len(df)
        min_col = 1
        max_col = len(df.columns)
        cells = self.sheet.range(min_row, min_col, max_row, max_col)

        gsheet_epoc = datetime.datetime(1899, 12, 30)
        for cell in cells:
            if cell.row == starting_row:
                cell.value = list(df.columns)[cell.col - 1]
            else:
                value = df.values[cell.row - starting_row - 1, cell.col - 1]
                # This is a hack.  Dates will be exported as days from epoch
                # Format this as date to get the dates you want
                if isinstance(value, pd.Timestamp):
                    value = (value.to_pydatetime() - gsheet_epoc).days
                elif value == "":
                    value = None
                elif type(value) in [np.int64, np.int32, np.int16, np.int8]:
                    value = int(value)
                cell.value = value
        self.sheet.update_cells(cells)



class DistFitter:
    def __init__(self, dist_class, num_hist_bins=None, hist_range=None):
        """
        Args:
            dist_class: A scipy.stats distribution class to fit to
            num_hist_bins: The number of bins to use for the empirical fitter
            hist_range: The range to use in the impirical fitter
        """
        self.dist_class = dist_class
        self._num_hist_bins = num_hist_bins
        self._hist_range = hist_range

    def fit(self, data, **fit_kwargs):
        """
        Args:
            data: iterable of data you want to fit to
        Kwargs:
            shape: The initial guess for the shape parameter(s) you want to pass to the fitter
            f0, f1, ...: Pin the shape parameter(s) to these values
            loc: The starting guess for the location parameter
            scale: the starting guess for the scale parameter
            floc: pin the lcoation parameter to this value
            fscale: pin the scale parameter to this value
        """
        from scipy import stats
        import numpy as np

        data = np.array(data)
        shape = fit_kwargs.pop("shape", None)
        if shape:
            args = (data, shape)
        else:
            args = (data,)
        self.dist_params = self.dist_class.fit(*args, **fit_kwargs)
        self.dist = self.dist_class(*self.dist_params)
        if self._num_hist_bins is None:
            hist_bins = int(len(data) / 5)
        else:
            hist_bins = self._num_hist_bins

        self.hist_dist = stats.rv_histogram(
            np.histogram(data, bins=hist_bins, range=self._hist_range)
        )
        min_val, max_val = self.hist_dist.ppf([0.01, 0.99])
        if self._hist_range:
            min_val, max_val = self._hist_range
        self.x = np.linspace(min_val, max_val, 600)

        (self.q_theory, self.q_data), (self._a, self._b, _) = stats.probplot(
            data, dist=self.dist, plot=None, fit=True
        )

        return self

    def _plot(
        self,
        label,
        method_name,
        logx=False,
        logy=False,
        data_color=None,
        fit_color=None,
        xlabel="value",
    ):
        import holoviews as hv
        import easier as ezr

        if data_color is None:
            data_color = ezr.cc[0]
        if fit_color is None:
            fit_color = ezr.cc[1]
        func = getattr(self.dist, method_name)
        hist_func = getattr(self.hist_dist, method_name)
        c1 = hv.Curve(
            (self.x, 1e-9 + func(self.x)),
            xlabel,
            "Density",
            label=f"{self.dist.dist.name} Fit",
        ).options(logx=logx, logy=logy, color=fit_color)
        c2 = hv.Curve(
            (self.x, 1e-9 + hist_func(self.x)), label=f"{label} Empirical"
        ).options(color=data_color)
        return hv.Overlay([c1, c2]).options(legend_position="top")

    def plot_pdf(
        self,
        label="",
        logx=False,
        logy=False,
        data_color=None,
        fit_color=None,
        xlabel="value",
    ):
        """
        Plot the pdf
        """
        return self._plot(
            label,
            method_name="pdf",
            logx=logx,
            logy=logy,
            data_color=data_color,
            fit_color=fit_color,
            xlabel=xlabel,
        )

    def plot_cdf(
        self,
        label="",
        logx=False,
        logy=False,
        data_color=None,
        fit_color=None,
        xlabel="value",
    ):
        """
        Plot the cdf
        """
        return self._plot(
            label,
            method_name="cdf",
            logx=logx,
            logy=logy,
            data_color=data_color,
            fit_color=fit_color,
            xlabel=xlabel,
        )

    def plot_qq(
        self, label="", logx=False, logy=False, data_color=None, fit_color=None
    ):
        """
        Make a q-q plot.  X axis will hold theoretical quantiles, Y axis the impirircal
        """
        import holoviews as hv
        import numpy as np

        if data_color is None:
            data_color = "blue"
        if fit_color is None:
            fit_color = "red"

        q_data_fit = np.polyval([self._a, self._b], self.q_theory)
        c1 = hv.Scatter(
            (self.q_theory, self.q_data),
            "Theoretical Quantiles",
            "Empirical Quantiles",
            label=label,
        ).options(size=5, alpha=0.5, color=data_color, logx=logx, logy=logy)
        label = f'{self.dist.dist.name}  params = {["%0.4g" % p for p in self.dist_params]}'.replace(
            "'", ""
        )
        c2 = hv.Curve((self.q_theory, q_data_fit), label="Best Fit Line").options(
            color=fit_color, alpha=0.2
        )
        c3 = hv.Curve((self.q_theory, self.q_theory), label=label).options(
            color="green", alpha=0.2
        )
        return hv.Overlay([c1, c2, c3]).options(legend_position="top")



from textwrap import dedent


class examples:
    """
    A descriptor whose only purpose is to print help text
    """

    def __get__(self, *args, **kwargs):
        print(
            dedent(
                """
            from easier import Fit

            #=================================================================
            #                Simple Model Fitting:  y ~ model(x)
            #=================================================================
            # Make data from noise-corrupted sinusoid
            x = np.linspace(0, 2 * np.pi, 100)
            y = np.sin(x) - .7 * np.cos(x) + .1 * np.random.randn(len(x))

            # Make up some artificial weights to apply to fit
            w = np.exp(8 * y)


            # Define a model function you want to fit to
            # All model parameters are on the p object.
            # The names "x", and "y" are reserved for the data you are fitting
            def model(p):
                return p.a * np.sin(p.k * p.x) + p.b * np.cos(p.k * p.x)

            # Initialize a fitter with purposefully bad guesses
            fitter = ezr.Fitter(a=-1, b=2, k=.2)

            # Fit the data with weights and plot fit quality every 5 iterations
            fitter.fit(x=x, y=y, weights=w, model=model, plot_every=5)

            # Plot the final results
            display(fitter.plot())
            display(fitter.params.df)

            #=================================================================
            #                Advanced Fitting:  y ~ model(x)
            #=================================================================
            # Make data from noise-corrupted sinusoid
            x = np.linspace(0, 2 * np.pi, 100)
            y = np.sin(x) - .7 * np.cos(x) + .1 * np.random.randn(len(x))

            # Define a model function you want to fit to
            # All model parameters are on the p object.
            # The names "x", and "y" are reserved for the data you are fitting
            def model(p):
                return p.a * np.sin(p.k * p.x) + p.b * np.cos(p.k * p.x)

            # Initialize a fitter with purposefully bad guesses
            fitter = ezr.Fitter(a=-1, b=2, k=.2)

            # Fit the data and plot fit quality every 5 iterations
            fitter.fit(
                x=x,                   # The independent data
                y=y,                   # The dependent data
                model=model,           # The model function
                plot_every=5,          # Plot fit every this number of iterations
                algorithm='fmin_bfgs', # Scipy optimization routine to use
                verbose=False          # Don't print convergence info
            )

            # Get predictions at specific values
            x_predict = np.linspace(0, 6 * np.pi, 300)
            y_predict = fitter.predict(x_predict)

            # Get the components of the fit chart
            components = fitter.plot(
                x=x_predict,
                scale_factor=10,
                label='10X Scaled Fit',
                line_color='red',
                scatter_color='blue',
                size=15,
                xlabel='My X Label',
                ylabel='My Y Label',
                as_components=True,
            )

            # Display the components as a layout rather than overlay
            display(hv.Layout(components))

            #=================================================================
            #                Advanced Fitting:  minimize: cost(x, y, p)
            #=================================================================
            # Assume we have x, y that are a third order polymial
            x_data = np.linspace(0, 1, 100)
            y_data = .7 * x_data ** 3 + .2 * x_data ** 2 + .1 * x_data
            y_data = y_data + .01 * np.random.randn(len(y_data))

            # Say that we know for sure that the function polyomial we want
            # has a zero intercept.  Furthermore, say we know for sure that
            # the sum of the coefficients have to equal to 1.  We construct
            # a cost function to encode this.
            def cost(p):
                # A full polynomial fit with all orders (so they can be thrown into polyfit)
                yf = p.a * p.x ** 3 + p.b * p.x ** 2 + p.c * p.x + p.d

                # Compute the sum of squares error cost of the fit from the data
                fit_cost = np.sum((yf - p.y)**2)

                # Regularize the length of the parameters to be one, and the intercept term to be zero.
                param_cost = 100 * len(p.x) * (p.a + p.b + p.c - 1) ** 2 + 10000 * len(p.x) * p.d ** 2

                # Return the total cost
                return fit_cost + param_cost

            # Create a fitter by initializing the the params to starting positions
            f = ezr.Fitter(a=1/3., b=1/3., c=1/3., d=0.)

            # Run the fit
            f.fit(x=x_data, y=y_data, cost=cost)
            p = f.params

            # Print the results
            print(f.params)
            """
            )
        )
        return None


class Fitter:
    """
    This class is a convenience wrapper for scipy optimization.
    Look at Fitter.examples attribute to see usage.

    The constructor takes exactly the same arguments as ParamState.
    """

    OPTIMIZER_NAMES = {"fmin", "fmin_powell", "fmin_cg", "fmin_bfgs"}

    DEFAULT_VERBOSE = True

    examples = examples()

    def __init__(self, *args, **kwargs):
        """
        *args and **kwargs define/initialize the model variables.
        They are passed directly to a ParamState constructor
        """
        from easier import ParamState

        self._params = ParamState(*args, **kwargs)
        self._algorithm = "fmin"
        self._optimizer_kwargs = {}
        self._model = None
        self._givens = {}
        self._cost = self._default_cost
        self._verbose = self.DEFAULT_VERBOSE

    @property
    def all_params(self):
        """
        Show all parameters (even data variables)
        """
        p = self._params.clone()
        return p

    @property
    def params(self):
        """
        Show all params except data variables
        """
        p = self.all_params
        p.drop("x")
        p.drop("y")
        return p

    def extra(self, **kwargs):
        """
        Put extra attributes on the params object.  These params will
        be completely ignored by the optimizer.  This is good for passing
        utility functions to your model or cost fuctions.
        """
        for key, val in kwargs.items():
            setattr(self._params, key, val)

    def given(self, **kwargs):
        """
        Supply kwargs that will set constant variables.  These can either
        be new variables or already defined variables.  This allows you
        to easily turn on and off variables to optimize.
        """
        self._givens = kwargs
        return self

    def optimizer_kwargs(self, **kwargs):
        """
        Additional kwargs to pass to optimizer.
        """
        self._optimizer_kwargs = kwargs
        return self

    def _cost_wrapper(self, args, p):
        """
        Allows users to define their cost functions as only functions
        of p, the ParamState object.
        """
        p.ingest(args)
        return self._cost(p)

    def _default_cost(self, p):
        """
        Allows users to define their cost functions as only functions
        of p, the ParamState object.
        """
        import numpy as np

        if hasattr(p, "weights"):
            w = p.weights
        else:
            w = np.ones_like(p.y)

        z = (self._model(p) - p.y) * w / np.sum(w)
        return np.sum(z**2)

    def _plotter(self, *args, **kwargs):
        """
        A function to plot fits in real time
        """
        import holoviews as hv
        import easier as ezr

        if kwargs["data"]:
            xd, yd = kwargs["data"][-2:]

        else:
            xd, yd = [0], [0]

        fit_line = hv.Curve(*args, **kwargs)
        data = hv.Scatter((xd, yd))
        overlay = hv.Overlay([fit_line, data])

        overlay = overlay.opts(
            hv.opts.Curve(color=ezr.cc.b),
            hv.opts.Scatter(color=ezr.cc.a, size=5, alpha=0.5),
        )
        return overlay

    def _model_wrapper(self, model):
        """
        The model wrapper will wrap the model with a function
        that updates real time plots of fit if needed.
        """
        if model is None:
            return model

        if self.plot_every is None:
            return model
        else:
            self.plot_counter = 0

            def wrapped(*args, **kwargs):
                yfit = model(*args, **kwargs)
                if self.plot_counter % self.plot_every == 0:
                    self.pipe.send(
                        (self._params.x, yfit, self._params.x, self._params.y)
                    )
                self.plot_counter += 1
                return yfit

            return wrapped

    def fit(
        self,
        *,
        x=None,
        y=None,
        weights=None,
        model=None,
        cost=None,
        plot_every=None,
        algorithm="fmin",
        verbose=True,
    ):
        """
        The method to use for training the fitter.
        Args:
                     x: the indepenant data
                     y: the depenant data
                 model: a model function to fit the data to
                  cost: a custom cost function (defaults to least squares)
            plot_every: Plot solution in real time every this number of iterations
             algorithm: Scipy optimization routine to use.  Enter nonsense to see list of valid.
               verbose: Print convergence information
        """
        import numpy as np
        from scipy import optimize
        import holoviews as hv
        from holoviews.streams import Pipe
        from IPython.display import display

        self.plot_every = plot_every

        if algorithm not in self.OPTIMIZER_NAMES:
            raise ValueError(
                f"Invalid optimizer {algorithm}.  Choose one of {self.OPTIMIZER_NAMES}"
            )

        givens = dict(x=x, y=y)
        if weights is not None:
            givens.update({"weights": weights})
        givens.update(self._givens)

        self._params.given(**givens)

        # This stuff only needs to happen if we are iteratively plotting fits
        if plot_every is not None:
            x, y = self._params.x, self._params.y
            xmin, xmax = np.min(x), np.max(x)
            ymin, ymax = np.min(y), np.max(y)

            scale = 0.1
            delta_x = scale * (xmax - xmin)
            delta_y = scale * (ymax - ymin)

            xmin, xmax = xmin - delta_x, xmax + delta_x
            ymin, ymax = ymin - delta_y, ymax + delta_y

            xlim = (xmin, xmax)
            ylim = (ymin, ymax)
            self.pipe = Pipe(data=[])
            try:
                dmap = hv.DynamicMap(self._plotter, streams=[self.pipe])
                dmap.opts(hv.opts.Overlay(xlim=xlim, ylim=ylim))
                display(dmap)
            except AttributeError:
                raise RuntimeError(
                    "You must import holoviews and set bokeh backround for plotting to work"
                )

        if model is None and cost is None:
            raise ValueError(
                "You must supply either a model function or a cost function"
            )

        self._raw_model = model
        self._model = self._model_wrapper(self._raw_model)

        a0 = self._params.array

        if cost is not None:
            self._cost = cost

        optimizer = getattr(optimize, algorithm)
        self._optimizer_kwargs.update(disp=verbose)
        a_fit = optimizer(
            self._cost_wrapper, a0, args=(self._params,), **self._optimizer_kwargs
        )
        a_fit = np.array(a_fit, ndmin=1)
        self._params.ingest(a_fit)
        return self

    def predict(self, x=None):
        """
        Returns an array of predictions based on trained models.  If
        x is not supplied, than the fit over x values in training set is used
        """
        if x is not None:
            p = self.all_params
            p.x = x
        else:
            p = self._params
        return self._raw_model(p)

    def df(self, x=None):
        import pandas as pd

        p = self._params

        if x is None:
            x = p.x

        y = self.predict(x)
        return pd.DataFrame({"x": x, "y": y})

    def plot(
        self,
        *,
        x=None,
        scale_factor=1,
        label=None,
        line_color=None,
        scatter_color=None,
        size=10,
        xlabel="x",
        ylabel="y",
        as_components=False,
    ):
        """
        Draw plots for model fit results.
        Params:
            x: A custom x over which to draw fits
            scale_factor: Scale all y values by this factor
            label: A string with which to label the fit
            line_color: Color for fit line
            scatter_color: Color for data points
            size: size of scatter points
            xlabel: x axis label
            ylabel: y axis label
            as_components: if True, return chart components rather than overlay
        """
        import easier as ezr

        p = self._params

        if x is None:
            x = p.x

        line_color = line_color if line_color else ezr.cc.b
        scatter_color = scatter_color if scatter_color else ezr.cc.a

        import holoviews as hv

        label_val = label if label else "Fit"

        try:
            scatter = hv.Scatter(
                (p.x, scale_factor * p.y), xlabel, ylabel, label=label_val
            ).options(color=scatter_color, size=size, alpha=0.5)
        except Exception:
            raise RuntimeError(
                "You must import holoviews and set bokeh backround for plotting to work"
            )

        line = hv.Curve((x, scale_factor * self.predict(x)), label=label_val).options(
            color=line_color
        )
        traces = [scatter, line]
        if as_components:
            return traces
        else:
            return hv.Overlay(traces)


# I'm not sure where this should actually live, but I'm putting it here for now.
def classifier_evaluation_plots(
    trained_model,
    X_test,
    y_test,
    threshold=0.5,
    plots="confusion_matrix, variable_importance, roc_curve, precision_recall, lift",
):
    """
    Creates classifier evalutaion plots.
    Args:
        trained_model: A trained sklearn or xbg model
        X_test: A test dataframe of X variables
        y_test A test dataframe of y variables (0 or 1)
        threshold: The operating point of the classifier
    """
    import matplotlib.pyplot as plt
    import numpy as np
    import pandas as pd
    from IPython.display import display
    from sklearn.metrics import (
        auc,
        average_precision_score,
        confusion_matrix,
        precision_recall_curve,
        roc_curve,
    )

    if "variable_importance" in plots:
        # Plot variable importance
        imp = pd.Series(
            trained_model.feature_importances_, index=trained_model.feature_names_in_
        ).sort_values()
        plt.figure()
        ax = imp.plot.barh(xlabel="Variable Importance")
        ax.figure.subplots_adjust(left=0.3)
        plt.show()

    # Step 2: Prediction and Threshold Selection
    y_pred_prob = trained_model.predict_proba(X_test)[
        :, 1
    ]  # probabilities of positive class
    y_pred = (y_pred_prob >= threshold).astype(int)

    if "confusion_matrix" in plots:
        # Step 3: Generating and Displaying Confusion Matrix
        conf_matrix = confusion_matrix(y_test, y_pred)
        conf_matrix_df = pd.DataFrame(
            conf_matrix,
            index=["True Negative", "True Positive"],
            columns=["Predicted Negative", "Predicted Positive"],
        )
        conf_matrix_df["Total"] = conf_matrix_df.sum(axis=1)
        conf_matrix_df = conf_matrix_df.T
        conf_matrix_df["Total"] = conf_matrix_df.sum(axis=1)
        conf_matrix_df = conf_matrix_df.T

        conf_matrix_df.index.name = "Counts"
        # print(conf_matrix_df.to_string())
        display(conf_matrix_df)
        print()

        dfcr = conf_matrix_df.divide(conf_matrix_df.loc[:, "Total"], axis=0)
        dfcr.index.name = "RowNorm"
        # print(dfcr.to_string())
        display(dfcr)
        print()

        dfcr = conf_matrix_df.divide(conf_matrix_df.loc["Total", :], axis=1)
        dfcr.index.name = "ColNorm"
        # print(dfcr.to_string())
        display(dfcr)

    if "roc_curve" in plots:
        # Step 4: Plotting ROC Curve and Calculating AUC
        fpr, tpr, thresholds_roc = roc_curve(y_test, y_pred_prob)
        roc_auc = auc(fpr, tpr)
        # Find the closest threshold to the selected one on the ROC curve
        closest_threshold_index = np.argmin(np.abs(thresholds_roc - threshold))
        operating_point_roc = (
            fpr[closest_threshold_index],
            tpr[closest_threshold_index],
        )

        plt.figure()
        plt.plot(
            fpr,
            tpr,
            color="darkorange",
            lw=2,
            label=f"ROC curve (area = {roc_auc:.2f})",
        )
        plt.plot([0, 1], [0, 1], color="navy", lw=2, linestyle="-")
        max_ind = min([len(fpr), len(thresholds_roc)])
        plt.plot(
            fpr[:max_ind],
            thresholds_roc[:max_ind],
            label="threshold (y ax)",
            color="grey",
            linestyle="--",
        )

        plt.scatter(*operating_point_roc, color="red", label="Operating Point")
        plt.xlabel("False Positive Rate")
        plt.ylabel("True Positive Rate")
        plt.title("Receiver Operating Characteristic (ROC)")
        plt.legend(loc="best")
        plt.show()

    if "precision_recall" in plots:
        # Step 5: Plotting Precision-Recall Curve
        precision, recall, thresholds_pr = precision_recall_curve(y_test, y_pred_prob)
        average_precision = average_precision_score(y_test, y_pred_prob)
        # Find the closest threshold to the selected one on the precision-recall curve
        closest_threshold_index = np.argmin(np.abs(thresholds_pr - threshold))
        operating_point_pr = (
            recall[closest_threshold_index],
            precision[closest_threshold_index],
        )

        plt.figure()
        plt.step(
            recall, precision, color="b", alpha=0.2, where="post", label="precision"
        )
        plt.fill_between(recall, precision, alpha=0.2, color="b", step="post")
        plt.scatter(*operating_point_pr, color="red", label="Operating Point")
        max_ind = min([len(recall), len(thresholds_pr)])
        plt.plot(
            recall[:max_ind],
            thresholds_pr[:max_ind],
            label="threshold (y ax)",
            color="grey",
            linestyle="--",
        )
        plt.xlabel(f"Recall = {np.round(recall[closest_threshold_index], 3)}")
        plt.ylabel(f"Precision = {np.round(precision[closest_threshold_index], 3)}")
        plt.ylim([0.0, 1.05])
        plt.xlim([0.0, 1.0])
        plt.title(f"Precision-Recall curve: AP={average_precision:.3f}")
        plt.legend(loc="best")
        plt.show()

    if "lift" in plots:
        lift = precision / precision[0]
        operating_point_lift = (
            recall[closest_threshold_index],
            lift[closest_threshold_index],
        )
        plt.figure()
        plt.plot(
            recall,
            lift,
            color="k",
        )
        plt.scatter(*operating_point_lift, color="red", label="Operating Point")

        plt.xlabel("Recall")
        plt.ylabel("Lift")
        plt.show()



import os
import dataclasses
import sys
import re
import time


class DuckMirror:
    """
    Mirrors a postgres database to a duckdb file.
    Authentication to the postgres database is read
    from the psql-friendly environment variables.
    """

    @dataclasses.dataclass
    class TableRec:
        source: str
        target: str

    def __init__(self, duckdb_file, overwrite=True):
        """
        Create a DuckMirror object
        Args:
            duckdb_file: The file to write the duckdb database to
            overwrite: If True, the file will be deleted if it exists
        """

        file = os.path.realpath(os.path.expanduser(duckdb_file))
        self.file_name = file
        self.overwrite = overwrite

    def clean(self):
        """
        A utility method to delete the duckdb file
        :return:
        """
        if os.path.isfile(self.file_name):
            os.unlink(self.file_name)

    def get_table_names(self):
        """
        Get the table names from the postgres database
        Returns: a list of TableRec objects
        """
        import duckdb

        conn = duckdb.connect()
        conn.execute("install postgres; load postgres;")
        conn.execute("ATTACH '' AS pg (TYPE POSTGRES, READ_ONLY);")

        df = conn.execute("use pg; show all tables").df()
        records = []
        for tup in df.itertuples():
            target = tup.name
            source = ".".join([tup.database, tup.schema, tup.name])
            records.append(self.TableRec(source=source, target=target))
        del conn
        return records

    def clone_tables(
        self,
        including=None,
        excluding=None,
        including_regex=None,
        excluding_regex=None,
        dry_run=False,
    ):
        """
        Clone the tables from the postgres database to the duckdb file
        Args:
            including: List[str] limit cloning to this  list of target table names
            excluding: List[str] exclude these target table names
        """
        import duckdb
        import pandas as pd

        if including is None:
            including = []
        if excluding is None:
            excluding = []

        if self.overwrite and not dry_run:
            self.clean()
        conn = duckdb.connect(self.file_name)
        try:
            conn.execute("install postgres; load postgres;")
            conn.execute("ATTACH '' AS pg (TYPE POSTGRES, READ_ONLY);")
        except duckdb.BinderException:
            pass

        recs = self.get_table_names()
        if including:
            recs = [r for r in recs if r.target in including]
        if excluding:
            recs = [r for r in recs if r.target not in excluding]

        if including_regex is not None:
            rex = re.compile(including_regex)
            recs = [r for r in recs if rex.match(r.target)]

        if excluding_regex is not None:
            rex = re.compile(excluding_regex)
            recs = [r for r in recs if not rex.match(r.target)]

        nrecs = len(recs)
        then = time.time()
        log_recs = []
        total_time = 0
        for nn, rec in enumerate(recs):
            print(f"{nn}/{nrecs}  {rec.target} : sec=", end="")
            sys.stdout.flush()
            sql = f"DROP TABLE IF EXISTS {rec.target}; CREATE TABLE {rec.target} AS FROM {rec.source};"
            if not dry_run:
                conn.execute(sql)
            now = time.time()
            delta = now - then
            then = now
            print(f"{delta:0.3f}")
            log_recs.append({"target": rec.target, "seconds": delta})
            total_time += delta
        print(f"\n----\nTotal time = {total_time:0.3} seconds")
        df = pd.DataFrame(log_recs)
        df = df.sort_values(by="seconds", ascending=False)
        return df



# This file is generated by Cog. Do not edit directly.
# Run 'cog -r tools.py' to regenerate.

# [[[cog
try:
    # In the easier subdirectory run cog -r tools.py to regenerate this file.
    import cog
    from easier.api import API, Importable

    for k, v in API.__dict__.items():
        if isinstance(v, Importable):
            cog.outl(f"from {v._module_path} import {v._artifact_name} as {k}")
            if v.instantiate:
                cog.outl(f"{k} = {k}()")
    del API
    del Importable
    del cog
    del k
    del v
except ImportError:
    pass
# ]]]
from easier.hvtools import Animator as Animator
from easier.bernstein import Bernstein as Bernstein
from easier.bernstein import BernsteinFitter as BernsteinFitter
from easier.utils import BlobAttr as BlobAttr
from easier.utils import BlobMixin as BlobMixin
from easier.clock import Clock as Clock
from easier.crypt import Crypt as Crypt
from easier.distributions import DistFitter as DistFitter
from easier.duck import Duck as Duck
from easier.duck_model import DuckModel as DuckModel
from easier.duck_mirror import DuckMirror as DuckMirror
from easier.filtering.elliptic_filter import Elliptic as Elliptic
from easier.fit_lib import Fitter as Fitter
from easier.gemini import Gemini as Gemini
from easier.clock import GlobalClock as GlobalClock
from easier.gsheet import GSheet as Gsheet
from easier.item import Item as Item
from easier.metabase import Metabase as Metabase
from easier.minimodel import MiniModel as MiniModel
from easier.minimodel import MiniModelPG as MiniModelPG
from easier.minimodel import MiniModelSqlite as MiniModelSqlite
from easier.ibis_tools import get_order_schema_class as OrderedSchema
OrderedSchema = OrderedSchema()
from easier.postgres import PG as PG
from easier.parallel import Parallel as Parallel
from easier.param_state import ParamState as ParamState
from easier.print_catcher import PrintCatcher as PrintCatcher
from easier.proportion_lib import Proportion as Proportion
from easier.dataframe_tools import get_quick_schema_class as QuickSchema
QuickSchema = QuickSchema()
from easier.dataframe_tools import get_pandas_sql_class as PandasSql
PandasSql = PandasSql()
from easier.salesforce import SalesForceReport as SalesForceReport
from easier.utils import Scaler as Scaler
from easier.shaper import Shaper as Shaper
from easier.salesforce import Soql as Soql
from easier.timer import Timer as Timer
from easier.hvtools import beta_plots as beta_plots
from easier.utils import cached_container as cached_container
from easier.utils import cached_dataframe as cached_dataframe
from easier.utils import cached_property as cached_property
from easier.plotting import ColorCyle as cc
cc = cc()
from easier.fit_lib import classifier_evaluation_plots as classifier_evaluation_plots
from easier.dataframe_tools import column_level_flattener as column_level_flattener
from easier.nice_dates import nice_dates as date_formatter
from easier.utils import diff_strings as diff_strings
from easier.utils import django_reconnect as django_reconnect
from easier.ecdf_lib import ecdf as ecdf
from easier.dataframe_tools import events_from_starting_ending as events_from_starting_ending
from easier.plotting import figure as figure
from easier.utils import get_logger as get_logger
from easier.dataframe_tools import heatmap as heatmap
from easier.dataframe_tools import hex_from_dataframe as hex_from_dataframe
from easier.dataframe_tools import hex_from_duckdb as hex_from_duckdb
from easier.dataframe_tools import hex_to_dataframe as hex_to_dataframe
from easier.dataframe_tools import hex_to_duckdb as hex_to_duckdb
from easier.hvtools import hist as hist
from easier.ibis_tools import ibis_conn_to_sqlalchemy_conn as ibis_conn_to_sqlalchemy_conn
from easier.ibis_tools import ibis_duck_connection as ibis_duck_connection
from easier.ibis_tools import get_sql as ibis_get_sql
from easier.ibis_tools import ibis_postgres_connection as ibis_postgres_connection
from easier.ibis_tools import sql_to_frame as ibis_sql_to_frame
from easier.utils import in_notebook as in_notebook
from easier.iterify import iterify as iterify
from easier.outlier_tools import kill_outliers_iqr as kill_outliers_iqr
from easier.outlier_tools import kill_outliers_sigma_edit as kill_outliers_sigma_edit
from easier.dataframe_tools import localize_utc_to_timezone as localize_utc_to_timezone
from easier.lomb_scargle import lomb_scargle as lomb_scargle
from easier.memory import mem_get as mem_get
from easier.memory import mem_show as mem_show
from easier.dataframe_tools import month_string as month_string
from easier.utils import mute_warnings as mute_warnings
from easier.nice_dates import nice_dates as nice_dates
from easier.normal_sampler import NormalSampleJoiner as NormalSampleJoiner
from easier.dataframe_tools import pandas_time_to_utc_seconds as pandas_time_to_utc_seconds
from easier.dataframe_tools import pandas_utc_seconds_to_time as pandas_utc_seconds_to_time
from easier.postgres import pg_creds_from_env as pg_creds_from_env
from easier.utils import pickle_cache_mixin as pickle_cache_mixin
from easier.utils import pickle_cache_state as pickle_cache_state
from easier.utils import pickle_cached_container as pickle_cached_container
from easier.utils import print_error as print_error
from easier.utils import python_type as python_type
from easier.utils import screen_width_full as screen_width_full
from easier.dataframe_tools import slugify as slugify
from easier.postgres import sql_file_to_df as sql_file_to_df
from easier.postgres import sql_string_to_df as sql_string_to_df
from easier.stream_plotter import StreamPlotter as StreamPlotter
from easier.utils import tqdm_flex as tqdm_flex
from easier.tracker import Tracker as Tracker
from easier.filtering.tvd import tvd as tvd
from easier.vonmises import VonMisesFitter as VonMisesFitter
from easier.utils import mute_warnings as warnings_mute
from easier.dataframe_tools import weekday_string as weekday_string
# [[[end]]]



class ListShaper:
    def flatten(self, shaped):
        import pandas as pd

        return pd.Series(shaped)

    def expand(self, flattened):
        return list(flattened)


class ArrayShaper:
    shape = None

    def flatten(self, shaped):
        import pandas as pd

        self.shape = shaped.shape
        return pd.Series(shaped.flatten())

    def expand(self, flattened):
        return flattened.values.reshape(self.shape)


class SeriesShaper:
    index = None

    def flatten(self, shaped):
        import pandas as pd

        self.index = shaped.index
        return pd.Series(shaped.values)

    def expand(self, flattened):
        import pandas as pd

        return pd.Series(
            flattened.values,
            index=self.index,
        )


class FrameShaper:
    index = None
    columns = None
    shape = None

    def flatten(self, shaped):
        import pandas as pd

        self.index = shaped.index
        self.columns = shaped.columns
        self.shape = shaped.values.shape

        return pd.Series(shaped.values.flatten())

    def expand(self, flattened):
        import pandas as pd

        return pd.DataFrame(
            flattened.values.reshape(self.shape), index=self.index, columns=self.columns
        )


class Shaper:
    """
    Datasets come in a variety of containers and sometimes you
    want to pass the data to a function that requires a one-dimenional
    container (usually an array).

    This class provides an easy method of transforming your data to
    and from a 1-dimentional series and back, preserving the containers
    shape (and columns/indexes for pandas objects.)

    The "flattened" data will always be a pandas series.

    The call signature is the same for all allowed datatypes.

    # Create a shaper
    shaper = ezr.Shaper()

    # Flatten the data
    flat_series = shaper.flatten(df)

    # Mutate the elements of the flat series
    # ** CAN NOT CHANGE THE LENGTH OR POSITION OF ELEMEMNTS **
    flat_series = process(flat_series)

    # Reconstitude the flat series to its original container
    # with possibly mutated elements
    df = shaper.expand(flat_series)

    Args:
        None: This class takes no arguments in its constructor.
    """

    def __init__(self):
        import numpy as np
        import pandas as pd

        # This holds an instance of the appropriate shaper
        self.shaper = None

        # This hold a mapping of allowed types to the appropriate shaper class
        self.type_mapper = {
            list: ListShaper,
            np.ndarray: ArrayShaper,
            pd.Series: SeriesShaper,
            pd.DataFrame: FrameShaper,
        }

    def flatten(self, shaped):
        """
        Flatten a data container to a pandas Series
        """
        # Make sure only allowed types are used
        if type(shaped) not in self.type_mapper:
            msg = f"Allowed input types are {list(self.type_mapper.keys())}"
            raise ValueError(msg)

        # Create an instance of the proper shaper
        self.shaper = self.type_mapper[type(shaped)]()

        # Returned the flattened data
        return self.shaper.flatten(shaped)

    def expand(self, flattened):
        # Return the expanded data
        return self.shaper.expand(flattened)



from typing import Optional, Tuple
from .utils import BlobAttr, BlobMixin, Scaler


class Compress:
    """A utility class that maps numbers into the range [0, 1].

    This class provides methods to compress and expand numerical values between a specified range
    and the [0, 1] interval. It can either learn the range from input data or use manually
    specified limits.

    Args:
        min_val (float, optional): Manually set the minimum value for the range. Defaults to None.
        max_val (float, optional): Manually set the maximum value for the range. Defaults to None.
    """

    def __init__(self, min_val=None, max_val=None):
        self._min_val = min_val
        self._max_val = max_val
        self.xmin = None
        self.ymin = None
        self._set_limits()

    def _set_limits(self, x=None):
        """Set the minimum and maximum values for the compression range.

        If x is provided, the limits are learned from the data. Otherwise, uses manually
        specified limits if they were provided during initialization.

        Args:
            x (numpy.ndarray, optional): Input data to learn limits from. Defaults to None.
        """
        import numpy as np

        xmin, xmax = None, None

        if x is not None:
            xmin = np.min(x)
            xmax = np.max(x)

        if self._min_val is not None:
            xmin = self._min_val
        if self._max_val is not None:
            xmax = self._max_val

        self.xmin = xmin
        self.xmax = xmax

    def compress(self, x, learn_limits=True):
        """Compress input values to the [0, 1] range.

        Args:
            x (numpy.ndarray): Input array to compress.
            learn_limits (bool, optional): Whether to learn the range limits from the input data.
                Defaults to True.

        Returns:
            numpy.ndarray: Compressed values in the range [0, 1].

        Raises:
            ValueError: If input is not a numpy array or if range limits are not set.
        """
        import numpy as np

        if not isinstance(x, np.ndarray):
            raise ValueError("Input must be numpy array")

        if learn_limits:
            self._set_limits(x)

        if None in {self.xmin, self.xmax}:
            raise ValueError("Compressor never learned range limit")

        return (x - self.xmin) / (self.xmax - self.xmin)

    def expand(self, x):
        """Expand values from [0, 1] range back to the original range.

        Args:
            x (numpy.ndarray): Input array in range [0, 1] to expand.

        Returns:
            numpy.ndarray: Expanded values in the original range.
        """
        x = x.flatten()
        xmin = self.xmin
        xmax = self.xmax

        x = xmin + (xmax - xmin) * x
        return x

    @property
    def _dc_dx(self):
        """Get the derivative of the compression function.

        Returns:
            float: The derivative of the compression function, which is the range width.
        """
        return self.xmax - self.xmin


class Bernstein(Compress):
    """Bernstein polynomial function approximator.

    This class implements a Bernstein polynomial approximation for either a callable function
    or x, y data points. It inherits from Compress to handle value normalization.

    Args:
        x (numpy.ndarray): x data points to fit.
        y (numpy.ndarray): y data points to fit.
        N (int, optional): Order of the fitting polynomial. Defaults to 500.
        xlim (tuple[float, float], optional): Hardcoded x limits within which to fit the function.
            Defaults to None.
    """

    _EPS = 1e-15

    def __init__(self, *, x, y, N=500, xlim: Optional[Tuple[float]] = None):
        if xlim is None:
            min_val, max_val = None, None
        else:
            min_val, max_val = xlim
        super().__init__(min_val, max_val)
        self._validate_inputs(x, y)
        self.N = N

        self._x = self.compress(x)
        self._y = y

        self._func = self._get_interp_function()

    def _get_interp_function(self):
        """Create an interpolation function from the input data.

        Returns:
            scipy.interpolate.interp1d: Interpolation function that handles out-of-bounds values
                by using the first and last y values.
        """
        from scipy.interpolate import interp1d

        return interp1d(
            self._x, self._y, fill_value=(self._y[0], self._y[-1]), bounds_error=False
        )

    def _validate_inputs(self, x, y):
        """Validate the input data for the Bernstein approximation.

        Args:
            x (numpy.ndarray): x data points.
            y (numpy.ndarray): y data points.

        Raises:
            ValueError: If inputs are not numpy arrays or have insufficient length.
        """
        import numpy as np

        if not all(isinstance(v, np.ndarray) for v in [x, y]):
            raise ValueError("x and y must both be numpy arrays")

        if not all(len(v) > 2 for v in [x, y]):
            raise ValueError("x and y must both have at least 3 elements")

    def _bern_term(self, n, k, x):
        """Calculate the Bernstein polynomial term using logarithms for numerical stability.

        This function uses logs to make the calculation safe for large n values.

        Args:
            n (int): Degree of the polynomial.
            k (int): Index of the term.
            x (numpy.ndarray): Input values.

        Returns:
            numpy.ndarray: Bernstein polynomial term values.
        """
        import numpy as np
        from scipy.special import gammaln

        x = np.clip(x, self._EPS, 1 - self._EPS)

        out = gammaln(n + 1) - gammaln(n - k + 1) - gammaln(k + 1)
        out += k * np.log(x) + (n - k) * np.log(1 - x)
        return np.exp(out)

    def _get_fit_func(self, infunc):
        """Create a Bernstein polynomial approximation function.

        Args:
            infunc (callable): Function to approximate.

        Returns:
            callable: Function that evaluates the Bernstein polynomial approximation.
        """
        import numpy as np

        N = self.N
        k_vec = np.arange(N + 1)
        coeff_vec = infunc(k_vec / N)

        def bern_sum(x):
            x = x.flatten()
            X, K = np.meshgrid(x, k_vec)
            _, C = np.meshgrid(x, coeff_vec)
            B = self._bern_term(N, K, X)

            terms = C * B
            out = np.sum(terms, axis=0)
            return out

        return bern_sum

    def _get_fit_deriv(self, infunc):
        """Create a Bernstein polynomial derivative approximation function.

        Args:
            infunc (callable): Function to approximate.

        Returns:
            callable: Function that evaluates the derivative of the Bernstein polynomial approximation.
        """
        import numpy as np

        N = self.N
        k_vec = np.arange(N + 1)
        coeff_vec = infunc(k_vec / N)

        def bern_sum(x):
            x = x.flatten()
            X, K = np.meshgrid(x, k_vec)
            _, C = np.meshgrid(x, coeff_vec)

            B1 = self._bern_term(N - 1, K - 1, X)
            B2 = self._bern_term(N - 1, K, X)

            terms = C * N * (B1 - B2)
            out = np.sum(terms, axis=0)
            return out

        return bern_sum

    def predict(self, x):
        """Evaluate the Bernstein polynomial approximation at given points.

        Args:
            x (numpy.ndarray): Points at which to evaluate the approximation.

        Returns:
            numpy.ndarray: Approximated function values at the input points.
        """
        func = self._get_fit_func(self._func)
        return func(self.compress(x))

    def predict_derivative(self, x):
        """Evaluate the derivative of the Bernstein polynomial approximation at given points.

        Args:
            x (numpy.ndarray): Points at which to evaluate the derivative approximation.

        Returns:
            numpy.ndarray: Approximated derivative values at the input points.
        """
        func = self._get_fit_deriv(self._func)
        out = func(self.compress(x))
        out = out / self._dc_dx
        return out


class BernsteinFitter(BlobMixin):
    """A class for fitting Bernstein polynomials to data with various constraints.

    This class implements a Bernstein polynomial fitter that can handle various constraints
    such as non-negativity, monotonicity, and endpoint matching. It inherits from BlobMixin
    for serialization capabilities.

    Args:
        non_negative (bool, optional): If True, constrains the fitted function to be non-negative.
            Defaults to False.
        monotonic (bool, optional): If True, constrains the fitted function to be monotonic.
            Defaults to False.
        increasing (bool, optional): If True and monotonic=True, constrains the function to be
            increasing. If False and monotonic=True, constrains it to be decreasing.
            Defaults to True.
        match_left (bool, optional): If True, matches the left endpoint value to the first data point.
            Defaults to False.
        match_right (bool, optional): If True, matches the right endpoint value to the last data point.
            Defaults to False.
        match_endpoint_values (bool, optional): If True, forces the function values at both endpoints
            to be equal. Cannot be used with match_left or match_right. Defaults to False.
        match_endpoint_derivatives (bool, optional): If True, forces the derivatives at both endpoints
            to be equal. Defaults to False.
    """

    _EPS = 1e-15

    w = BlobAttr(None)
    scaler_blob = BlobAttr(None)

    def __init__(
        self,
        non_negative=False,
        monotonic=False,
        increasing=True,
        match_left=False,
        match_right=False,
        match_endpoint_values=False,
        match_endpoint_derivatives=False,
    ):
        super().__init__()
        self._non_negative = non_negative
        self._monotonic = monotonic
        self._increasing = increasing
        self._match_left = match_left
        self._match_right = match_right
        self._match_endpoint_values = match_endpoint_values
        self._match_endpoint_derivatives = match_endpoint_derivatives

    def _bern_term(self, n, k, x):
        """Calculate a single Bernstein polynomial term using logarithms for numerical stability.

        This function uses logs to make the calculation safe for large n values.

        Args:
            n (int): Degree of the polynomial.
            k (int): Index of the term.
            x (numpy.ndarray): Input values.

        Returns:
            numpy.ndarray: Bernstein polynomial term values.
        """
        import numpy as np
        from scipy.special import gammaln

        if (k < 0) or (k > n):
            return np.zeros_like(x)

        x = np.clip(x, self._EPS, 1 - self._EPS)

        out = gammaln(n + 1) - gammaln(n - k + 1) - gammaln(k + 1)
        out += k * np.log(x) + (n - k) * np.log(1 - x)
        return np.exp(out)

    def _get_design_matrix(self, x, degree):
        """Create the design matrix for Bernstein polynomial fitting.

        Args:
            x (numpy.ndarray): Array of x points to evaluate the polynomials at.
            degree (int): Degree of the Bernstein polynomial.

        Returns:
            numpy.ndarray: Design matrix of shape (len(x), degree + 1).
        """
        import numpy as np

        x = np.array(x)
        A = np.zeros((len(x), degree + 1))
        for k in range(0, degree + 1):
            A[:, k] = self._bern_term(degree, k, x)
        return A

    def _get_derivative_matrix(self, x, degree):
        """Create the derivative matrix for Bernstein polynomial fitting.

        Args:
            x (numpy.ndarray): Array of x points to evaluate the derivatives at.
            degree (int): Degree of the Bernstein polynomial.

        Returns:
            numpy.ndarray: Derivative matrix of shape (len(x), degree + 1).
        """
        import numpy as np

        n = degree
        if hasattr(x, "__iter__"):
            B = np.zeros((len(x), degree + 1))
        else:
            B = np.zeros((1, degree + 1))

        for k in range(0, degree + 1):
            term1 = self._bern_term(n - 1, k - 1, x)
            term2 = self._bern_term(n - 1, k, x)
            B[:, k] = n * (term1 - term2)
        return B

    def _get_integral_matrix(self, x, degree):
        """Create the integral matrix for Bernstein polynomial fitting.

        Args:
            x (numpy.ndarray): Array of x points to evaluate the integrals at.
            degree (int): Degree of the Bernstein polynomial.

        Returns:
            numpy.ndarray: Integral matrix of shape (len(x), degree + 1).
        """
        import numpy as np

        n = degree

        A = self._get_design_matrix(x, degree + 1)
        B = np.zeros_like(A)
        coeff = 1 / (n + 1)
        for k in range(0, degree + 1):
            X = A[:, (k + 1) :]
            B[:, k] = coeff * np.sum(X, axis=1)
        return B

    def get_design_matrix(self, x, degree):
        """Get the design matrix with scaled input values.

        Args:
            x (numpy.ndarray): Array of x points to evaluate the polynomials at.
            degree (int): Degree of the Bernstein polynomial.

        Returns:
            numpy.ndarray: Design matrix of shape (len(x), degree + 1).
        """
        import numpy as np

        x = np.array(x)

        scaler = Scaler()
        x = scaler.fit_transform(x)
        self.scaler_blob = scaler.to_blob()

        A = self._get_design_matrix(x, degree)
        return A

    def fit_predict(self, x, y, degree, regulizer=0.0, verbose=False):
        """Fit the Bernstein polynomial and return predictions for the input points.

        Args:
            x (numpy.ndarray): Array of x points to fit.
            y (numpy.ndarray): Array of y values to fit.
            degree (int): Degree of the Bernstein polynomial.
            regulizer (float, optional): Regularization strength. Defaults to 0.0.
            verbose (bool, optional): Whether to print optimization progress. Defaults to False.

        Returns:
            numpy.ndarray: Predicted values at the input points.
        """
        self.fit(x, y, degree, regulizer=regulizer, verbose=verbose)
        return self.predict(x)

    def fit(self, x, y, degree, sample_weights=None, regulizer=0.0, verbose=False):
        """Fit a Bernstein polynomial to the given data with specified constraints.

        Args:
            x (numpy.ndarray): Array of x points to fit.
            y (numpy.ndarray): Array of y values to fit.
            degree (int): Degree of the Bernstein polynomial.
            sample_weights (numpy.ndarray, optional): Array of weights for each data point. Defaults to None.
            regulizer (float, optional): Regularization strength. Defaults to 0.0.
            verbose (bool, optional): Whether to print optimization progress. Defaults to False.

        Returns:
            BernsteinFitter: The fitted instance for method chaining.

        Raises:
            ValueError: If match_endpoint_values is used with match_left or match_right.
        """
        import cvxpy as cp
        import numpy as np

        x = np.array(x)
        y = np.array(y)

        scaler = Scaler()
        x = scaler.fit_transform(x)
        self.scaler_blob = scaler.to_blob()

        yv = np.reshape(y, (-1, 1))
        A = self._get_design_matrix(x, degree)
        B = self._get_derivative_matrix(x, degree)

        # Define a weight variable to be optimized
        w = cp.Variable(name="w", shape=(degree + 1, 1))

        # The objective is the mininum squared error
        # objective = cp.Minimize(cp.sum_squares(A @ w - yv) + regulizer * cp.norm(w, 2))
        if sample_weights is None:
            objective = cp.Minimize(
                cp.sum_squares(A @ w - yv) + regulizer * cp.norm(w, 2)
            )
        else:
            sample_weights = np.array(sample_weights)
            # objective = cp.Minimize(
            #     cp.sum(sample_weights * cp.square(A @ w - yv))
            #     + regulizer * cp.norm(w, 2)
            # )
            objective = cp.Minimize(
                cp.sum(cp.multiply(sample_weights, cp.square(A @ w - yv)))
                + regulizer * cp.norm(w, 2)
            )

        # Default to unconstrained
        constraints = []

        if self._non_negative:
            constraints.append(w >= np.zeros(w.shape))

        if self._monotonic:
            if self._increasing:
                constraints.append(B @ w >= np.zeros_like(yv))
            else:
                constraints.append(B @ w <= np.zeros_like(yv))

        if self._match_left:
            constraints.append(w[0, 0] == y[0])

        if self._match_right:
            constraints.append(w[-1, 0] == y[-1])

        # These constraints ensure matching values at end points (periodic)
        if self._match_endpoint_values:
            if self._match_left or self._match_right:
                raise ValueError(
                    "Cannot have match_left or match_right with match_endpoint_values"
                )
            # Constrain the values at the endpoints to match
            constraints.append(w[0, 0] == w[-1, 0])

        # These constraints ensure matching derivative at end points (periodic)
        if self._match_endpoint_derivatives:
            # Get the derivative matrix
            B = self._get_derivative_matrix(x, degree)

            # Constrain the derivatives at the endpoints to match
            term1 = B[0, :] @ w
            term2 = B[-1, :] @ w
            constraints.append(term1 == term2)

        # Add any desired constraints
        kwargs = {}
        if constraints:
            kwargs["constraints"] = constraints

        # Solve the problem
        problem = cp.Problem(objective, **kwargs)
        problem.solve(verbose=verbose)

        self.w = w.value.flatten()

        return self

    def predict(self, x):
        """Predict values using the fitted Bernstein polynomial.

        Values that fall outside the fitted x range will be pegged to
        the terminal values of the fitter.

        Args:
            x (numpy.ndarray): Points at which to evaluate the fitted function.

        Returns:
            numpy.ndarray: Predicted values at the input points.

        Raises:
            ValueError: If fit() has not been called or no blob has been loaded.
        """
        return self._get_prediction(x, "value")

    def predict_derivative(self, x):
        """Predict derivatives using the fitted Bernstein polynomial.

        Args:
            x (numpy.ndarray): Points at which to evaluate the derivative.

        Returns:
            numpy.ndarray: Predicted derivatives at the input points.

        Raises:
            ValueError: If fit() has not been called or no blob has been loaded.
        """
        if self.w is None:
            raise ValueError(
                "You must run fit() or load a blob before running predict()"
            )

        scaler = Scaler()
        scaler.from_blob(self.scaler_blob)
        diffs = self._get_prediction(x, "derivative")
        return diffs / (scaler.limits[1] - scaler.limits[0])

    def predict_integral(self, x):
        """Predict integrals using the fitted Bernstein polynomial.

        Args:
            x (numpy.ndarray): Points at which to evaluate the integral.

        Returns:
            numpy.ndarray: Predicted integrals at the input points.

        Raises:
            ValueError: If fit() has not been called or no blob has been loaded.
        """
        if self.w is None:
            raise ValueError(
                "You must run fit() or load a blob before running predict()"
            )

        scaler = Scaler()
        scaler.from_blob(self.scaler_blob)
        result = self._get_prediction(x, "integral")
        return result * (scaler.limits[1] - scaler.limits[0])

    def _get_prediction(self, x, what):
        """Internal method to get predictions, derivatives, or integrals.

        Args:
            x (numpy.ndarray): Points at which to evaluate.
            what (str): Type of prediction to make ("value", "derivative", or "integral").

        Returns:
            numpy.ndarray: Predicted values at the input points.

        Raises:
            ValueError: If fit() has not been called or no blob has been loaded.
            ValueError: If what is not one of "value", "derivative", or "integral".
        """
        import numpy as np

        if self.w is None:
            raise ValueError(
                "You must run fit() or load a blob before running predict()"
            )

        is_scalar = False
        if not hasattr(x, "__iter__"):
            is_scalar = True
            x = [x]

        x = np.array(x)
        scaler = Scaler()
        scaler.from_blob(self.scaler_blob)
        x = scaler.transform(x)

        degree = len(self.w) - 1
        wv = np.reshape(self.w, (-1, 1))
        if what == "value":
            A = self._get_design_matrix(x, degree)
            yv = A @ wv
        elif what == "derivative":
            B = self._get_derivative_matrix(x, degree)
            yv = B @ wv
        elif what == "integral":
            B = self._get_integral_matrix(x, degree)
            yv = B @ wv
        else:
            raise ValueError(f'Nope!  {what!r} is a bad "what" argument')

        yv = yv.flatten()
        if is_scalar:
            return yv[0]
        else:
            return yv

    def to_blob(self):
        """Convert the fitter to a serializable blob.

        Returns:
            dict: A dictionary containing the fitter's state.
        """
        blob = super().to_blob()
        blob["w"] = list(blob["w"])
        return blob

    def from_blob(self, blob):
        """Load the fitter's state from a blob.

        Args:
            blob (dict): A dictionary containing the fitter's state.

        Returns:
            BernsteinFitter: The loaded instance for method chaining.
        """
        import numpy as np

        super().from_blob(blob)
        self.w = np.array(self.w)

        return self

    def get_polynomial_coefficients(self):
        """Convert Bernstein coefficients to standard polynomial coefficients.

        Returns coefficients in descending order of power to match numpy's polyfit/polyval convention.
        The coefficients are transformed to work with unscaled x values.

        Returns:
            numpy.ndarray: Polynomial coefficients in descending order of power.

        Raises:
            ValueError: If fit() has not been called or no blob has been loaded.

        Example:
            If result is [3, 2, 1], the polynomial is 3x^2 + 2x + 1
        """
        import numpy as np

        if self.w is None:
            raise ValueError(
                "You must run fit() or load a blob before getting coefficients"
            )

        degree = len(self.w) - 1

        # Create a set of points to evaluate at in scaled space
        x_scaled = np.linspace(0, 1, degree + 1)

        # Get the scaler to transform back to unscaled space
        scaler = Scaler()
        scaler.from_blob(self.scaler_blob)
        x_min, x_max = scaler.limits
        x_range = x_max - x_min

        # Transform scaled points to unscaled space
        x_unscaled = x_scaled * x_range + x_min

        # Create Vandermonde matrix using unscaled x values
        V = np.vander(x_unscaled, degree + 1, increasing=True)

        # Get y values using predict which handles scaling internally
        y = self.predict(x_unscaled)

        # Solve for polynomial coefficients
        poly_coeffs = np.linalg.solve(V, y)

        return poly_coeffs[::-1]  # Reverse to match numpy convention
